r"""
Collect information about software installed on Windows OS
================

:maintainer: Salt Stack <https://github.com/saltstack>
:codeauthor: Damon Atkins <https://github.com/damon-atkins>
:maturity: new
:depends: pywin32
:platform: windows

Known Issue: install_date may not match Control Panel\Programs\Programs and Features
"""
import collections
import datetime
import locale
import logging
import os.path
import platform
import re
import sys
import time
from functools import cmp_to_key

__version__ = "0.1"

try:
    import pywintypes
    import win32api
    import win32con
    import win32process
    import win32security
    import winerror

except ImportError:
    if __name__ == "__main__":
        raise ImportError("Please install pywin32/pypiwin32")
    else:
        raise


if __name__ == "__main__":
    LOG_CONSOLE = logging.StreamHandler()
    LOG_CONSOLE.setFormatter(logging.Formatter("[%(levelname)s]: %(message)s"))
    log = logging.getLogger(__name__)
    log.addHandler(LOG_CONSOLE)
    log.setLevel(logging.DEBUG)
else:
    log = logging.getLogger(__name__)


try:
    from salt.utils.odict import OrderedDict
except ImportError:
    from collections import OrderedDict

from salt.utils.versions import Version

# pylint: disable=too-many-instance-attributes


class RegSoftwareInfo:
    """
    Retrieve Registry data on a single installed software item or component.

    Attribute:
        None

    :codeauthor: Damon Atkins <https://github.com/damon-atkins>
    """

    # Variables shared by all instances
    __guid_pattern = re.compile(
        r"^\{(\w{8})-(\w{4})-(\w{4})-(\w\w)(\w\w)-(\w\w)(\w\w)(\w\w)(\w\w)(\w\w)(\w\w)\}$"
    )
    __squid_pattern = re.compile(
        r"^(\w{8})(\w{4})(\w{4})(\w\w)(\w\w)(\w\w)(\w\w)(\w\w)(\w\w)(\w\w)(\w\w)$"
    )
    __version_pattern = re.compile(r"\d+\.\d+\.\d+[\w.-]*|\d+\.\d+[\w.-]*")
    __upgrade_codes = {}
    __upgrade_code_have_scan = {}

    __reg_types = {
        "str": (win32con.REG_EXPAND_SZ, win32con.REG_SZ),
        "list": (win32con.REG_MULTI_SZ),
        "int": (win32con.REG_DWORD, win32con.REG_DWORD_BIG_ENDIAN, win32con.REG_QWORD),
        "bytes": (win32con.REG_BINARY),
    }

    # Search 64bit, on 64bit platform, on 32bit its ignored
    if platform.architecture()[0] == "32bit":
        # Handle Python 32bit on 64&32 bit platform and Python 64bit
        if win32process.IsWow64Process():  # pylint: disable=no-member
            # 32bit python on a 64bit platform
            __use_32bit_lookup = {True: 0, False: win32con.KEY_WOW64_64KEY}
        else:
            # 32bit python on a 32bit platform
            __use_32bit_lookup = {True: 0, False: None}
    else:
        __use_32bit_lookup = {True: win32con.KEY_WOW64_32KEY, False: 0}

    def __init__(self, key_guid, sid=None, use_32bit=False):
        """
        Initialise against a software item or component.

        All software has a unique "Identifer" within the registry. This can be free
        form text/numbers e.g. "MySoftware" or
        GUID e.g. "{0EAF0D8F-C9CF-4350-BD9A-07EC66929E04}"

        Args:
            key_guid (str): Identifer.
            sid (str): Security IDentifier of the User or None for Computer/Machine.
            use_32bit (bool):
                Regisrty location of the Identifer. ``True`` 32 bit registry only
                meaning fully on 64 bit OS.
        """
        self.__reg_key_guid = key_guid  # also called IdentifyingNumber(wmic)
        self.__squid = ""
        self.__reg_products_path = ""
        self.__reg_upgradecode_path = ""
        self.__patch_list = None

        # If a valid GUID create the SQUID also.
        guid_match = self.__guid_pattern.match(key_guid)
        if guid_match is not None:
            for index in range(1, 12):
                # __guid_pattern breaks up the GUID
                self.__squid += guid_match.group(index)[::-1]

        if sid:
            # User data seems to be more spreadout within the registry.
            self.__reg_hive = "HKEY_USERS"
            self.__reg_32bit = False  # Force to False
            self.__reg_32bit_access = (
                0  # HKEY_USERS does not have a 32bit and 64bit view
            )
            self.__reg_uninstall_path = "{}\\Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{}".format(
                sid, key_guid
            )
            if self.__squid:
                self.__reg_products_path = (
                    "{}\\Software\\Classes\\Installer\\Products\\{}".format(
                        sid, self.__squid
                    )
                )
                self.__reg_upgradecode_path = (
                    "{}\\Software\\Microsoft\\Installer\\UpgradeCodes".format(sid)
                )
                self.__reg_patches_path = (
                    "Software\\Microsoft\\Windows\\CurrentVersion\\Installer\\UserData\\"
                    "{}\\Products\\{}\\Patches".format(sid, self.__squid)
                )
        else:
            self.__reg_hive = "HKEY_LOCAL_MACHINE"
            self.__reg_32bit = use_32bit
            self.__reg_32bit_access = self.__use_32bit_lookup[use_32bit]
            self.__reg_uninstall_path = (
                "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{}".format(
                    key_guid
                )
            )
            if self.__squid:
                self.__reg_products_path = (
                    "Software\\Classes\\Installer\\Products\\{}".format(self.__squid)
                )
                self.__reg_upgradecode_path = (
                    "Software\\Classes\\Installer\\UpgradeCodes"
                )
                self.__reg_patches_path = (
                    "Software\\Microsoft\\Windows\\CurrentVersion\\Installer\\UserData\\"
                    "S-1-5-18\\Products\\{}\\Patches".format(self.__squid)
                )

        # OpenKey is expensive, open in advance and keep it open.
        # This must exist
        try:
            # pylint: disable=no-member
            self.__reg_uninstall_handle = win32api.RegOpenKeyEx(
                getattr(win32con, self.__reg_hive),
                self.__reg_uninstall_path,
                0,
                win32con.KEY_READ | self.__reg_32bit_access,
            )
        except pywintypes.error as exc:  # pylint: disable=no-member
            if exc.winerror == winerror.ERROR_FILE_NOT_FOUND:
                log.error(
                    "Software/Component Not Found  key_guid: '%s', "
                    "sid: '%s' , use_32bit: '%s'",
                    key_guid,
                    sid,
                    use_32bit,
                )
            raise  # This must exist or have no errors

        self.__reg_products_handle = None
        if self.__squid:
            try:
                # pylint: disable=no-member
                self.__reg_products_handle = win32api.RegOpenKeyEx(
                    getattr(win32con, self.__reg_hive),
                    self.__reg_products_path,
                    0,
                    win32con.KEY_READ | self.__reg_32bit_access,
                )
            except pywintypes.error as exc:  # pylint: disable=no-member
                if exc.winerror == winerror.ERROR_FILE_NOT_FOUND:
                    log.debug(
                        "Software/Component Not Found in Products section of registry "
                        "key_guid: '%s', sid: '%s', use_32bit: '%s'",
                        key_guid,
                        sid,
                        use_32bit,
                    )
                    self.__squid = None  # mark it as not a SQUID
                else:
                    raise

        self.__mod_time1970 = 0
        # pylint: disable=no-member
        mod_win_time = win32api.RegQueryInfoKeyW(self.__reg_uninstall_handle).get(
            "LastWriteTime", None
        )
        # pylint: enable=no-member
        if mod_win_time:
            # at some stage __int__() was removed from pywintypes.datetime to return secs since 1970
            if hasattr(mod_win_time, "utctimetuple"):
                self.__mod_time1970 = time.mktime(mod_win_time.utctimetuple())
            elif hasattr(mod_win_time, "__int__"):
                self.__mod_time1970 = int(mod_win_time)

    def __squid_to_guid(self, squid):
        """
        Squished GUID (SQUID) to GUID.

        A SQUID is a Squished/Compressed version of a GUID to use up less space
        in the registry.

        Args:
            squid (str): Squished GUID.

        Returns:
            str: the GUID if a valid SQUID provided.
        """
        if not squid:
            return ""
        squid_match = self.__squid_pattern.match(squid)
        guid = ""
        if squid_match is not None:
            guid = (
                "{"
                + squid_match.group(1)[::-1]
                + "-"
                + squid_match.group(2)[::-1]
                + "-"
                + squid_match.group(3)[::-1]
                + "-"
                + squid_match.group(4)[::-1]
                + squid_match.group(5)[::-1]
                + "-"
            )
            for index in range(6, 12):
                guid += squid_match.group(index)[::-1]
            guid += "}"
        return guid

    @staticmethod
    def __one_equals_true(value):
        """
        Test for ``1`` as a number or a string and return ``True`` if it is.

        Args:
            value: string or number or None.

        Returns:
            bool: ``True`` if 1 otherwise ``False``.
        """
        if isinstance(value, int) and value == 1:
            return True
        elif (
            isinstance(value, str)
            and re.match(r"\d+", value, flags=re.IGNORECASE + re.UNICODE) is not None
            and str(value) == "1"
        ):
            return True
        return False

    @staticmethod
    def __reg_query_value(handle, value_name):
        """
        Calls RegQueryValueEx

        If PY2 ensure unicode string and expand REG_EXPAND_SZ before returning
        Remember to catch not found exceptions when calling.

        Args:
            handle (object): open registry handle.
            value_name (str): Name of the value you wished returned

        Returns:
            tuple: type, value
        """
        # item_value, item_type = win32api.RegQueryValueEx(self.__reg_uninstall_handle, value_name)
        item_value, item_type = win32api.RegQueryValueEx(
            handle, value_name
        )  # pylint: disable=no-member
        if item_type == win32con.REG_EXPAND_SZ:
            # expects Unicode input
            win32api.ExpandEnvironmentStrings(item_value)  # pylint: disable=no-member
            item_type = win32con.REG_SZ
        return item_value, item_type

    @property
    def install_time(self):
        """
        Return the install time, or provide an estimate of install time.

        Installers or even self upgrading software must/should update the date
        held within InstallDate field when they change versions. Some installers
        do not set ``InstallDate`` at all so we use the last modified time on the
        registry key.

        Returns:
            int: Seconds since 1970 UTC.
        """
        time1970 = self.__mod_time1970  # time of last resort
        try:
            # pylint: disable=no-member
            date_string, item_type = win32api.RegQueryValueEx(
                self.__reg_uninstall_handle, "InstallDate"
            )
        except pywintypes.error as exc:  # pylint: disable=no-member
            if exc.winerror == winerror.ERROR_FILE_NOT_FOUND:
                return time1970  # i.e. use time of last resort
            else:
                raise

        if item_type == win32con.REG_SZ:
            try:
                date_object = datetime.datetime.strptime(date_string, "%Y%m%d")
                time1970 = time.mktime(date_object.timetuple())
            except ValueError:  # date format is not correct
                pass

        return time1970

    def get_install_value(self, value_name, wanted_type=None):
        """
        For the uninstall section of the registry return the name value.

        Args:
            value_name (str): Registry value name.
            wanted_type (str):
                The type of value wanted if the type does not match
                None is return. wanted_type support values are
                ``str`` ``int`` ``list`` ``bytes``.

        Returns:
            value: Value requested or None if not found.
        """
        try:
            item_value, item_type = self.__reg_query_value(
                self.__reg_uninstall_handle, value_name
            )
        except pywintypes.error as exc:  # pylint: disable=no-member
            if exc.winerror == winerror.ERROR_FILE_NOT_FOUND:
                # Not Found
                return None
            raise

        if wanted_type and item_type not in self.__reg_types[wanted_type]:
            item_value = None

        return item_value

    def is_install_true(self, key):
        """
        For the uninstall section check if name value is ``1``.

        Args:
            value_name (str): Registry value name.

        Returns:
            bool: ``True`` if ``1`` otherwise ``False``.
        """
        return self.__one_equals_true(self.get_install_value(key))

    def get_product_value(self, value_name, wanted_type=None):
        """
        For the product section of the registry return the name value.

        Args:
            value_name (str): Registry value name.
            wanted_type (str):
                The type of value wanted if the type does not match
                None is return. wanted_type support values are
                ``str`` ``int`` ``list`` ``bytes``.

        Returns:
            value: Value requested or ``None`` if not found.
        """
        if not self.__reg_products_handle:
            return None
        subkey, search_value_name = os.path.split(value_name)
        try:
            if subkey:

                handle = win32api.RegOpenKeyEx(  # pylint: disable=no-member
                    self.__reg_products_handle,
                    subkey,
                    0,
                    win32con.KEY_READ | self.__reg_32bit_access,
                )
                item_value, item_type = self.__reg_query_value(
                    handle, search_value_name
                )
                win32api.RegCloseKey(handle)  # pylint: disable=no-member
            else:
                item_value, item_type = win32api.RegQueryValueEx(
                    self.__reg_products_handle, value_name
                )  # pylint: disable=no-member
        except pywintypes.error as exc:  # pylint: disable=no-member
            if exc.winerror == winerror.ERROR_FILE_NOT_FOUND:
                # Not Found
                return None
            raise

        if wanted_type and item_type not in self.__reg_types[wanted_type]:
            item_value = None
        return item_value

    @property
    def upgrade_code(self):
        """
        For installers which follow the Microsoft Installer standard, returns
        the ``Upgrade code``.

        Returns:
            value (str): ``Upgrade code`` GUID for installed software.
        """
        if not self.__squid:
            # Must have a valid squid for an upgrade code to exist
            return ""

        # GUID/SQUID are unique, so it does not matter if they are 32bit or
        # 64bit or user install so all items are cached into a single dict
        have_scan_key = "{}\\{}\\{}".format(
            self.__reg_hive, self.__reg_upgradecode_path, self.__reg_32bit
        )
        if not self.__upgrade_codes or self.__reg_key_guid not in self.__upgrade_codes:
            # Read in the upgrade codes in this section of the registry.
            try:
                uc_handle = win32api.RegOpenKeyEx(
                    getattr(win32con, self.__reg_hive),  # pylint: disable=no-member
                    self.__reg_upgradecode_path,
                    0,
                    win32con.KEY_READ | self.__reg_32bit_access,
                )
            except pywintypes.error as exc:  # pylint: disable=no-member
                if exc.winerror == winerror.ERROR_FILE_NOT_FOUND:
                    # Not Found
                    log.warning(
                        "Not Found %s\\%s 32bit %s",
                        self.__reg_hive,
                        self.__reg_upgradecode_path,
                        self.__reg_32bit,
                    )
                    return ""
                raise
            squid_upgrade_code_all, _, _, suc_pytime = zip(
                *win32api.RegEnumKeyEx(uc_handle)
            )  # pylint: disable=no-member

            # Check if we have already scanned these upgrade codes before, and also
            # check if they have been updated in the registry since last time we scanned.
            if (
                have_scan_key in self.__upgrade_code_have_scan
                and self.__upgrade_code_have_scan[have_scan_key]
                == (
                    squid_upgrade_code_all,
                    suc_pytime,
                )
            ):
                log.debug(
                    "Scan skipped for upgrade codes, no changes (%s)", have_scan_key
                )
                return ""  # we have scanned this before and no new changes.

            # Go into each squid upgrade code and find all the related product codes.
            log.debug("Scan for upgrade codes (%s) for product codes", have_scan_key)
            for upgrade_code_squid in squid_upgrade_code_all:
                upgrade_code_guid = self.__squid_to_guid(upgrade_code_squid)
                pc_handle = win32api.RegOpenKeyEx(
                    uc_handle,  # pylint: disable=no-member
                    upgrade_code_squid,
                    0,
                    win32con.KEY_READ | self.__reg_32bit_access,
                )
                _, pc_val_count, _ = win32api.RegQueryInfoKey(
                    pc_handle
                )  # pylint: disable=no-member
                for item_index in range(pc_val_count):
                    product_code_guid = self.__squid_to_guid(
                        win32api.RegEnumValue(pc_handle, item_index)[0]
                    )  # pylint: disable=no-member
                    if product_code_guid:
                        self.__upgrade_codes[product_code_guid] = upgrade_code_guid
                win32api.RegCloseKey(pc_handle)  # pylint: disable=no-member

            win32api.RegCloseKey(uc_handle)  # pylint: disable=no-member
            self.__upgrade_code_have_scan[have_scan_key] = (
                squid_upgrade_code_all,
                suc_pytime,
            )

        return self.__upgrade_codes.get(self.__reg_key_guid, "")

    @property
    def list_patches(self):
        """
        For installers which follow the Microsoft Installer standard, returns
        a list of patches applied.

        Returns:
            value (list): Long name of the patch.
        """
        if not self.__squid:
            # Must have a valid squid for an upgrade code to exist
            return []

        if self.__patch_list is None:
            # Read in the upgrade codes in this section of the reg.
            try:
                pat_all_handle = win32api.RegOpenKeyEx(
                    getattr(win32con, self.__reg_hive),  # pylint: disable=no-member
                    self.__reg_patches_path,
                    0,
                    win32con.KEY_READ | self.__reg_32bit_access,
                )
            except pywintypes.error as exc:  # pylint: disable=no-member
                if exc.winerror == winerror.ERROR_FILE_NOT_FOUND:
                    # Not Found
                    log.warning(
                        "Not Found %s\\%s 32bit %s",
                        self.__reg_hive,
                        self.__reg_patches_path,
                        self.__reg_32bit,
                    )
                    return []
                raise

            pc_sub_key_cnt, _, _ = win32api.RegQueryInfoKey(
                pat_all_handle
            )  # pylint: disable=no-member
            if not pc_sub_key_cnt:
                return []
            squid_patch_all, _, _, _ = zip(
                *win32api.RegEnumKeyEx(pat_all_handle)
            )  # pylint: disable=no-member

            ret = []
            # Scan the patches for the DisplayName of active patches.
            for patch_squid in squid_patch_all:
                try:
                    patch_squid_handle = (
                        win32api.RegOpenKeyEx(  # pylint: disable=no-member
                            pat_all_handle,
                            patch_squid,
                            0,
                            win32con.KEY_READ | self.__reg_32bit_access,
                        )
                    )
                    (
                        patch_display_name,
                        patch_display_name_type,
                    ) = self.__reg_query_value(patch_squid_handle, "DisplayName")
                    patch_state, patch_state_type = self.__reg_query_value(
                        patch_squid_handle, "State"
                    )
                    if (
                        patch_state_type != win32con.REG_DWORD
                        or not isinstance(patch_state_type, int)
                        or patch_state != 1
                        or patch_display_name_type  # 1 is Active, 2 is Superseded/Obsolute
                        != win32con.REG_SZ
                    ):
                        continue
                    win32api.RegCloseKey(
                        patch_squid_handle
                    )  # pylint: disable=no-member
                    ret.append(patch_display_name)
                except pywintypes.error as exc:  # pylint: disable=no-member
                    if exc.winerror == winerror.ERROR_FILE_NOT_FOUND:
                        log.debug("skipped patch, not found %s", patch_squid)
                        continue
                    raise

        return ret

    @property
    def registry_path_text(self):
        """
        Returns the uninstall path this object is associated with.

        Returns:
            str: <hive>\\<uninstall registry entry>
        """
        return "{}\\{}".format(self.__reg_hive, self.__reg_uninstall_path)

    @property
    def registry_path(self):
        """
        Returns the uninstall path this object is associated with.

        Returns:
            tuple: hive, uninstall registry entry path.
        """
        return (self.__reg_hive, self.__reg_uninstall_path)

    @property
    def guid(self):
        """
        Return GUID or Key.

        Returns:
            str: GUID or Key
        """
        return self.__reg_key_guid

    @property
    def squid(self):
        """
        Return SQUID of the GUID if a valid GUID.

        Returns:
            str: GUID
        """
        return self.__squid

    @property
    def package_code(self):
        """
        Return package code of the software.

        Returns:
            str: GUID
        """
        return self.__squid_to_guid(self.get_product_value("PackageCode"))

    @property
    def version_binary(self):
        """
        Return version number which is stored in binary format.

        Returns:
            str: <major 0-255>.<minior 0-255>.<build 0-65535> or None if not found
        """
        # Under MSI 'Version' is a 'REG_DWORD' which then sets other registry
        # values like DisplayVersion to x.x.x to the same value.
        # However not everyone plays by the rules, so we need to check first.
        # version_binary_data will be None if the reg value does not exist.
        # Some installs set 'Version' to REG_SZ (string) which is not
        # the MSI standard
        try:
            item_value, item_type = self.__reg_query_value(
                self.__reg_uninstall_handle, "version"
            )
        except pywintypes.error as exc:  # pylint: disable=no-member
            if exc.winerror == winerror.ERROR_FILE_NOT_FOUND:
                # Not Found
                return "", ""

        version_binary_text = ""
        version_src = ""
        if item_value:
            if item_type == win32con.REG_DWORD:
                if isinstance(item_value, int):
                    version_binary_raw = item_value
                if version_binary_raw:
                    # Major.Minor.Build
                    version_binary_text = "{}.{}.{}".format(
                        version_binary_raw >> 24 & 0xFF,
                        version_binary_raw >> 16 & 0xFF,
                        version_binary_raw & 0xFFFF,
                    )
                    version_src = "binary-version"

            elif (
                item_type == win32con.REG_SZ
                and isinstance(item_value, str)
                and self.__version_pattern.match(item_value) is not None
            ):
                # Hey, version should be a int/REG_DWORD, an installer has set
                # it to a string
                version_binary_text = item_value.strip(" ")
                version_src = "binary-version (string)"

        return (version_binary_text, version_src)


class WinSoftware:
    """
    Point in time snapshot of the software and components installed on
    a system.

    Attributes:
        None

    :codeauthor: Damon Atkins <https://github.com/damon-atkins>
    """

    __sid_pattern = re.compile(r"^S-\d-\d-\d+$|^S-\d-\d-\d+-\d+-\d+-\d+-\d+$")
    __whitespace_pattern = re.compile(r"^\s*$", flags=re.UNICODE)
    # items we copy out of the uninstall section of the registry without further processing
    __uninstall_search_list = [
        ("url", "str", ["URLInfoAbout", "HelpLink", "MoreInfoUrl", "UrlUpdateInfo"]),
        ("size", "int", ["Size", "EstimatedSize"]),
        ("win_comments", "str", ["Comments"]),
        ("win_release_type", "str", ["ReleaseType"]),
        ("win_product_id", "str", ["ProductID"]),
        ("win_product_codes", "str", ["ProductCodes"]),
        ("win_package_refs", "str", ["PackageRefs"]),
        ("win_install_location", "str", ["InstallLocation"]),
        ("win_install_src_dir", "str", ["InstallSource"]),
        ("win_parent_pkg_uid", "str", ["ParentKeyName"]),
        ("win_parent_name", "str", ["ParentDisplayName"]),
    ]
    # items we copy out of the products section of the registry without further processing
    __products_search_list = [
        ("win_advertise_flags", "int", ["AdvertiseFlags"]),
        ("win_redeployment_flags", "int", ["DeploymentFlags"]),
        ("win_instance_type", "int", ["InstanceType"]),
        ("win_package_name", "str", ["SourceList\\PackageName"]),
    ]

    def __init__(self, version_only=False, user_pkgs=False, pkg_obj=None):
        """
        Point in time snapshot of the software and components installed on
        a system.

        Args:
            version_only (bool): Provide list of versions installed instead of detail.
            user_pkgs (bool): Include software/components installed with user space.
            pkg_obj (object):
                If None (default) return default package naming standard and use
                default version capture methods (``DisplayVersion`` then
                ``Version``, otherwise ``0.0.0.0``)
        """
        self.__pkg_obj = pkg_obj  # must be set before calling get_software_details
        self.__version_only = version_only
        self.__reg_software = {}
        self.__get_software_details(user_pkgs=user_pkgs)
        self.__pkg_cnt = len(self.__reg_software)
        self.__iter_list = None

    @property
    def data(self):
        """
        Returns the raw data

        Returns:
            dict: contents of the dict are dependent on the parameters passed
                when the class was initiated.
        """
        return self.__reg_software

    @property
    def version_only(self):
        """
        Returns True if class initiated with ``version_only=True``

        Returns:
            bool: The value of ``version_only``
        """
        return self.__version_only

    def __len__(self):
        """
        Returns total number of software/components installed.

        Returns:
            int: total number of software/components installed.
        """
        return self.__pkg_cnt

    def __getitem__(self, pkg_id):
        """
        Returns information on a package.

        Args:
            pkg_id (str): Package Id of the software/component

        Returns:
            dict or list: List if ``version_only`` is ``True`` otherwise dict
        """
        if pkg_id in self.__reg_software:
            return self.__reg_software[pkg_id]
        else:
            raise KeyError(pkg_id)

    def __iter__(self):
        """
        Standard interation class initialisation over package information.
        """
        if self.__iter_list is not None:
            raise RuntimeError("Can only perform one iter at a time")
        self.__iter_list = collections.deque(sorted(self.__reg_software.keys()))
        return self

    def __next__(self):
        """
        Returns next Package Id.

        Returns:
            str: Package Id
        """
        try:
            return self.__iter_list.popleft()
        except IndexError:
            self.__iter_list = None
            raise StopIteration

    def next(self):
        """
        Returns next Package Id.

        Returns:
            str: Package Id
        """
        return self.__next__()

    def get(self, pkg_id, default_value=None):
        """
        Returns information on a package.

        Args:
            pkg_id (str): Package Id of the software/component.
            default_value: Value to return when the Package Id is not found.

        Returns:
            dict or list: List if ``version_only`` is ``True`` otherwise dict
        """
        return self.__reg_software.get(pkg_id, default_value)

    @staticmethod
    def __oldest_to_latest_version(ver1, ver2):
        """
        Used for sorting version numbers oldest to latest
        """
        return 1 if Version(ver1) > Version(ver2) else -1

    @staticmethod
    def __latest_to_oldest_version(ver1, ver2):
        """
        Used for sorting version numbers, latest to oldest
        """
        return 1 if Version(ver1) < Version(ver2) else -1

    def pkg_version_list(self, pkg_id):
        """
        Returns information on a package.

        Args:
            pkg_id (str): Package Id of the software/component.

        Returns:
            list: List of version numbers installed.
        """
        pkg_data = self.__reg_software.get(pkg_id, None)
        if not pkg_data:
            return []

        if isinstance(pkg_data, list):
            # raw data is 'pkgid': [sorted version list]
            return pkg_data  # already sorted oldest to newest

        # Must be a dict or OrderDict, and contain full details
        installed_versions = list(pkg_data.get("version").keys())
        return sorted(
            installed_versions, key=cmp_to_key(self.__oldest_to_latest_version)
        )

    def pkg_version_latest(self, pkg_id):
        """
        Returns a package latest version installed out of all the versions
        currently installed.

        Args:
            pkg_id (str): Package Id of the software/component.

        Returns:
            str: Latest/Newest version number installed.
        """
        return self.pkg_version_list(pkg_id)[-1]

    def pkg_version_oldest(self, pkg_id):
        """
        Returns a package oldest version installed out of all the versions
        currently installed.

        Args:
            pkg_id (str): Package Id of the software/component.

        Returns:
            str: Oldest version number installed.
        """
        return self.pkg_version_list(pkg_id)[0]

    @staticmethod
    def __sid_to_username(sid):
        """
        Provided with a valid Windows Security Identifier (SID) and returns a Username

        Args:
            sid (str): Security Identifier (SID).

        Returns:
            str: Username in the format of username@realm or username@computer.
        """
        if sid is None or sid == "":
            return ""
        try:
            sid_bin = win32security.GetBinarySid(sid)  # pylint: disable=no-member
        except pywintypes.error as exc:  # pylint: disable=no-member
            raise ValueError(
                "pkg: Software owned by {} is not valid: [{}] {}".format(
                    sid, exc.winerror, exc.strerror
                )
            )
        try:
            name, domain, _account_type = win32security.LookupAccountSid(
                None, sid_bin
            )  # pylint: disable=no-member
            user_name = "{}\\{}".format(domain, name)
        except pywintypes.error as exc:  # pylint: disable=no-member
            # if user does not exist...
            # winerror.ERROR_NONE_MAPPED = No mapping between account names and
            # security IDs was carried out.
            if exc.winerror == winerror.ERROR_NONE_MAPPED:  # 1332
                # As the sid is from the registry it should be valid
                # even if it cannot be lookedup, so the sid is returned
                return sid
            else:
                raise ValueError(
                    "Failed looking up sid '{}' username: [{}] {}".format(
                        sid, exc.winerror, exc.strerror
                    )
                )
        try:
            user_principal = win32security.TranslateName(  # pylint: disable=no-member
                user_name,
                win32api.NameSamCompatible,  # pylint: disable=no-member
                win32api.NameUserPrincipal,
            )  # pylint: disable=no-member
        except pywintypes.error as exc:  # pylint: disable=no-member
            # winerror.ERROR_NO_SUCH_DOMAIN The specified domain either does not exist
            # or could not be contacted, computer may not be part of a domain also
            # winerror.ERROR_INVALID_DOMAINNAME The format of the specified domain name is
            # invalid. e.g. S-1-5-19 which is a local account
            # winerror.ERROR_NONE_MAPPED No mapping between account names and security IDs was done.
            if exc.winerror in (
                winerror.ERROR_NO_SUCH_DOMAIN,
                winerror.ERROR_INVALID_DOMAINNAME,
                winerror.ERROR_NONE_MAPPED,
            ):
                return "{}@{}".format(name.lower(), domain.lower())
            else:
                raise
        return user_principal

    def __software_to_pkg_id(self, publisher, name, is_component, is_32bit):
        """
        Determine the Package ID of a software/component using the
        software/component ``publisher``, ``name``, whether its a software or a
        component, and if its 32bit or 64bit archiecture.

        Args:
            publisher (str): Publisher of the software/component.
            name (str): Name of the software.
            is_component (bool): True if package is a component.
            is_32bit (bool): True if the software/component is 32bit architecture.

        Returns:
            str: Package Id
        """
        if publisher:
            # remove , and lowercase as , are used as list separators
            pub_lc = publisher.replace(",", "").lower()

        else:
            # remove , and lowercase
            pub_lc = "NoValue"  # Capitals/Special Value

        if name:
            name_lc = name.replace(",", "").lower()
            # remove ,   OR we do the URL Encode on chars we do not want e.g. \\ and ,
        else:
            name_lc = "NoValue"  # Capitals/Special Value

        if is_component:
            soft_type = "comp"
        else:
            soft_type = "soft"

        if is_32bit:
            soft_type += "32"  # Tag only the 32bit only

        default_pkg_id = pub_lc + "\\\\" + name_lc + "\\\\" + soft_type

        # Check to see if class was initialise with pkg_obj with a method called
        # to_pkg_id, and if so use it for the naming standard instead of the default
        if self.__pkg_obj and hasattr(self.__pkg_obj, "to_pkg_id"):
            pkg_id = self.__pkg_obj.to_pkg_id(publisher, name, is_component, is_32bit)
            if pkg_id:
                return pkg_id

        return default_pkg_id

    def __version_capture_slp(
        self, pkg_id, version_binary, version_display, display_name
    ):
        """
        This returns the version and where the version string came from, based on instructions
        under ``version_capture``, if ``version_capture`` is missing, it defaults to
        value of display-version.

        Args:
            pkg_id (str): Publisher of the software/component.
            version_binary (str): Name of the software.
            version_display (str): True if package is a component.
            display_name (str): True if the software/component is 32bit architecture.

        Returns:
            str: Package Id
        """
        if self.__pkg_obj and hasattr(self.__pkg_obj, "version_capture"):
            version_str, src, version_user_str = self.__pkg_obj.version_capture(
                pkg_id, version_binary, version_display, display_name
            )
            if src != "use-default" and version_str and src:
                return version_str, src, version_user_str
            elif src != "use-default":
                raise ValueError(
                    "version capture within object '{}' failed "
                    "for pkg id: '{}' it returned '{}' '{}' "
                    "'{}'".format(
                        str(self.__pkg_obj),
                        pkg_id,
                        version_str,
                        src,
                        version_user_str,
                    )
                )

        # If self.__pkg_obj.version_capture() not defined defaults to using
        # version_display and if not valid then use version_binary, and as a last
        # result provide the version 0.0.0.0.0 to indicate version string was not determined.
        if (
            version_display
            and re.match(r"\d+", version_display, flags=re.IGNORECASE + re.UNICODE)
            is not None
        ):
            version_str = version_display
            src = "display-version"
        elif (
            version_binary
            and re.match(r"\d+", version_binary, flags=re.IGNORECASE + re.UNICODE)
            is not None
        ):
            version_str = version_binary
            src = "version-binary"
        else:
            src = "none"
            version_str = "0.0.0.0.0"
        # return version str, src of the version, "user" interpretation of the version
        # which by default is version_str
        return version_str, src, version_str

    def __collect_software_info(self, sid, key_software, use_32bit):
        """
        Update data with the next software found
        """

        reg_soft_info = RegSoftwareInfo(key_software, sid, use_32bit)

        # Check if the registry entry is a valid.
        # a) Cannot manage software without at least a display name
        display_name = reg_soft_info.get_install_value("DisplayName", wanted_type="str")
        if display_name is None or self.__whitespace_pattern.match(display_name):
            return

        # b) make sure its not an 'Hotfix', 'Update Rollup', 'Security Update', 'ServicePack'
        # General this is software which pre dates Windows 10
        default_value = reg_soft_info.get_install_value("", wanted_type="str")
        release_type = reg_soft_info.get_install_value("ReleaseType", wanted_type="str")

        if (
            re.match(
                r"^{.*\}\.KB\d{6,}$", key_software, flags=re.IGNORECASE + re.UNICODE
            )
            is not None
            or (default_value and default_value.startswith(("KB", "kb", "Kb")))
            or (
                release_type
                and release_type
                in ("Hotfix", "Update Rollup", "Security Update", "ServicePack")
            )
        ):
            log.debug("skipping hotfix/update/service pack %s", key_software)
            return

        # if NoRemove exists we would expect their to be no UninstallString
        uninstall_no_remove = reg_soft_info.is_install_true("NoRemove")
        uninstall_string = reg_soft_info.get_install_value("UninstallString")
        uninstall_quiet_string = reg_soft_info.get_install_value("QuietUninstallString")
        uninstall_modify_path = reg_soft_info.get_install_value("ModifyPath")
        windows_installer = reg_soft_info.is_install_true("WindowsInstaller")
        system_component = reg_soft_info.is_install_true("SystemComponent")
        publisher = reg_soft_info.get_install_value("Publisher", wanted_type="str")

        # UninstallString is optional if the installer is "windows installer"/MSI
        # However for it to appear in Control-Panel -> Program and Features -> Uninstall or change a program
        # the UninstallString needs to be set or ModifyPath set
        if (
            uninstall_string is None
            and uninstall_quiet_string is None
            and uninstall_modify_path is None
            and (not windows_installer)
        ):
            return

        # Question: If uninstall string is not set and windows_installer should we set it
        # Question: if uninstall_quiet is not set .......

        if sid:
            username = self.__sid_to_username(sid)
        else:
            username = None

        # We now have a valid software install or a system component
        pkg_id = self.__software_to_pkg_id(
            publisher, display_name, system_component, use_32bit
        )
        version_binary, version_src = reg_soft_info.version_binary
        version_display = reg_soft_info.get_install_value(
            "DisplayVersion", wanted_type="str"
        )
        # version_capture is what the slp defines, the result overrides. Question: maybe it should error if it fails?
        (version_text, version_src, user_version) = self.__version_capture_slp(
            pkg_id, version_binary, version_display, display_name
        )
        if not user_version:
            user_version = version_text

        # log.trace('%s\\%s ver:%s src:%s', username or 'SYSTEM', pkg_id, version_text, version_src)

        if username:
            dict_key = "{};{}".format(
                username, pkg_id
            )  # Use ; as its not a valid hostnmae char
        else:
            dict_key = pkg_id

        # Guessing the architecture http://helpnet.flexerasoftware.com/isxhelp21/helplibrary/IHelp64BitSupport.htm
        # A 32 bit installed.exe can install a 64 bit app, but for it to write to 64bit reg it will
        # need to use WOW. So the following is a bit of a guess

        if self.__version_only:
            # package name and package version list, are the only info being return
            if dict_key in self.__reg_software:
                if version_text not in self.__reg_software[dict_key]:
                    # Not expecting the list to be big, simple search and insert
                    insert_point = 0
                    for ver_item in self.__reg_software[dict_key]:
                        if Version(version_text) <= Version(ver_item):
                            break
                        insert_point += 1
                    self.__reg_software[dict_key].insert(insert_point, version_text)
                else:
                    # This code is here as it can happen, especially if the
                    # package id provided by pkg_obj is simple.
                    log.debug(
                        "Found extra entries for '%s' with same version "
                        "'%s', skipping entry '%s'",
                        dict_key,
                        version_text,
                        key_software,
                    )
            else:
                self.__reg_software[dict_key] = [version_text]

            return

        if dict_key in self.__reg_software:
            data = self.__reg_software[dict_key]
        else:
            data = self.__reg_software[dict_key] = OrderedDict()

        if sid:
            # HKEY_USERS has no 32bit and 64bit view like HKEY_LOCAL_MACHINE
            data.update({"arch": "unknown"})
        else:
            arch_str = "x86" if use_32bit else "x64"
            if "arch" in data:
                if data["arch"] != arch_str:
                    data["arch"] = "many"
            else:
                data.update({"arch": arch_str})

        if publisher:
            if "vendor" in data:
                if data["vendor"].lower() != publisher.lower():
                    data["vendor"] = "many"
            else:
                data["vendor"] = publisher

        if "win_system_component" in data:
            if data["win_system_component"] != system_component:
                data["win_system_component"] = None
        else:
            data["win_system_component"] = system_component

        data.update({"win_version_src": version_src})

        data.setdefault("version", {})
        if version_text in data["version"]:
            if "win_install_count" in data["version"][version_text]:
                data["version"][version_text]["win_install_count"] += 1
            else:
                # This is only defined when we have the same item already
                data["version"][version_text]["win_install_count"] = 2
        else:
            data["version"][version_text] = OrderedDict()

        version_data = data["version"][version_text]
        version_data.update({"win_display_name": display_name})
        if uninstall_string:
            version_data.update({"win_uninstall_cmd": uninstall_string})
        if uninstall_quiet_string:
            version_data.update({"win_uninstall_quiet_cmd": uninstall_quiet_string})
        if uninstall_no_remove:
            version_data.update({"win_uninstall_no_remove": uninstall_no_remove})

        version_data.update({"win_product_code": key_software})
        if version_display:
            version_data.update({"win_version_display": version_display})
        if version_binary:
            version_data.update({"win_version_binary": version_binary})
        if user_version:
            version_data.update({"win_version_user": user_version})

        # Determine Installer Product
        #   'NSIS:Language'
        #   'Inno Setup: Setup Version'
        if windows_installer or (
            uninstall_string
            and re.search(
                r"MsiExec.exe\s|MsiExec\s",
                uninstall_string,
                flags=re.IGNORECASE + re.UNICODE,
            )
        ):
            version_data.update({"win_installer_type": "winmsi"})
        elif re.match(r"InstallShield_", key_software, re.IGNORECASE) is not None or (
            uninstall_string
            and (
                re.search(
                    r"InstallShield", uninstall_string, flags=re.IGNORECASE + re.UNICODE
                )
                is not None
                or re.search(
                    r"isuninst\.exe.*\.isu",
                    uninstall_string,
                    flags=re.IGNORECASE + re.UNICODE,
                )
                is not None
            )
        ):
            version_data.update({"win_installer_type": "installshield"})
        elif key_software.endswith("_is1") and reg_soft_info.get_install_value(
            "Inno Setup: Setup Version", wanted_type="str"
        ):
            version_data.update({"win_installer_type": "inno"})
        elif uninstall_string and re.search(
            r".*\\uninstall.exe|.*\\uninst.exe",
            uninstall_string,
            flags=re.IGNORECASE + re.UNICODE,
        ):
            version_data.update({"win_installer_type": "nsis"})
        else:
            version_data.update({"win_installer_type": "unknown"})

        # Update dict with information retrieved so far for detail results to be return
        # Do not add fields which are blank.
        language_number = reg_soft_info.get_install_value("Language")
        if (
            isinstance(language_number, int)
            and language_number in locale.windows_locale
        ):
            version_data.update(
                {"win_language": locale.windows_locale[language_number]}
            )

        package_code = reg_soft_info.package_code
        if package_code:
            version_data.update({"win_package_code": package_code})

        upgrade_code = reg_soft_info.upgrade_code
        if upgrade_code:
            version_data.update({"win_upgrade_code": upgrade_code})

        is_minor_upgrade = reg_soft_info.is_install_true("IsMinorUpgrade")
        if is_minor_upgrade:
            version_data.update({"win_is_minor_upgrade": is_minor_upgrade})

        install_time = reg_soft_info.install_time
        if install_time:
            version_data.update(
                {
                    "install_date": datetime.datetime.fromtimestamp(
                        install_time
                    ).isoformat()
                }
            )
            version_data.update({"install_date_time_t": int(install_time)})

        for infokey, infotype, regfield_list in self.__uninstall_search_list:
            for regfield in regfield_list:
                strvalue = reg_soft_info.get_install_value(
                    regfield, wanted_type=infotype
                )
                if strvalue:
                    version_data.update({infokey: strvalue})
                    break

        for infokey, infotype, regfield_list in self.__products_search_list:
            for regfield in regfield_list:
                data = reg_soft_info.get_product_value(regfield, wanted_type=infotype)
                if data is not None:
                    version_data.update({infokey: data})
                    break
        patch_list = reg_soft_info.list_patches
        if patch_list:
            version_data.update({"win_patches": patch_list})

    def __get_software_details(self, user_pkgs):
        """
        This searches the uninstall keys in the registry to find
        a match in the sub keys, it will return a dict with the
        display name as the key and the version as the value
        .. sectionauthor:: Damon Atkins <https://github.com/damon-atkins>
        .. versionadded:: 2016.11.0
        """

        # FUNCTION MAIN CODE #
        # Search 64bit, on 64bit platform, on 32bit its ignored.
        if platform.architecture()[0] == "32bit":
            # Handle Python 32bit on 64&32 bit platform and Python 64bit
            if win32process.IsWow64Process():  # pylint: disable=no-member
                # 32bit python on a 64bit platform
                use_32bit_lookup = {True: 0, False: win32con.KEY_WOW64_64KEY}
                arch_list = [True, False]
            else:
                # 32bit python on a 32bit platform
                use_32bit_lookup = {True: 0, False: None}
                arch_list = [True]

        else:
            # Python is 64bit therefore most be on 64bit System.
            use_32bit_lookup = {True: win32con.KEY_WOW64_32KEY, False: 0}
            arch_list = [True, False]

        # Process software installed for the machine i.e. all users.
        for arch_flag in arch_list:
            key_search = "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall"
            log.debug("SYSTEM processing 32bit:%s", arch_flag)
            handle = win32api.RegOpenKeyEx(  # pylint: disable=no-member
                win32con.HKEY_LOCAL_MACHINE,
                key_search,
                0,
                win32con.KEY_READ | use_32bit_lookup[arch_flag],
            )
            reg_key_all, _, _, _ = zip(
                *win32api.RegEnumKeyEx(handle)
            )  # pylint: disable=no-member
            win32api.RegCloseKey(handle)  # pylint: disable=no-member
            for reg_key in reg_key_all:
                self.__collect_software_info(None, reg_key, arch_flag)

        if not user_pkgs:
            return

        # Process software installed under all USERs, this adds significate processing time.
        # There is not 32/64 bit registry redirection under user tree.
        log.debug("Processing user software... please wait")
        handle_sid = win32api.RegOpenKeyEx(  # pylint: disable=no-member
            win32con.HKEY_USERS, "", 0, win32con.KEY_READ
        )
        sid_all = []
        for index in range(
            win32api.RegQueryInfoKey(handle_sid)[0]
        ):  # pylint: disable=no-member
            sid_all.append(
                win32api.RegEnumKey(handle_sid, index)
            )  # pylint: disable=no-member

        for sid in sid_all:
            if (
                self.__sid_pattern.match(sid) is not None
            ):  # S-1-5-18 needs to be ignored?
                user_uninstall_path = "{}\\Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall".format(
                    sid
                )
                try:
                    handle = win32api.RegOpenKeyEx(  # pylint: disable=no-member
                        handle_sid, user_uninstall_path, 0, win32con.KEY_READ
                    )
                except pywintypes.error as exc:  # pylint: disable=no-member
                    if exc.winerror == winerror.ERROR_FILE_NOT_FOUND:
                        # Not Found Uninstall under SID
                        log.debug("Not Found %s", user_uninstall_path)
                        continue
                    else:
                        raise
                try:
                    reg_key_all, _, _, _ = zip(
                        *win32api.RegEnumKeyEx(handle)
                    )  # pylint: disable=no-member
                except ValueError:
                    log.debug("No Entries Found %s", user_uninstall_path)
                    reg_key_all = []
                win32api.RegCloseKey(handle)  # pylint: disable=no-member
                for reg_key in reg_key_all:
                    self.__collect_software_info(sid, reg_key, False)
        win32api.RegCloseKey(handle_sid)  # pylint: disable=no-member
        return


def __main():
    """This module can also be run directly for testing
    Args:
        detail|list : Provide ``detail`` or version ``list``.
        system|system+user: System installed and System and User installs.
    """
    if len(sys.argv) < 3:
        sys.stderr.write(
            "usage: {} <detail|list> <system|system+user>\n".format(sys.argv[0])
        )
        sys.exit(64)
    user_pkgs = False
    version_only = False
    if str(sys.argv[1]) == "list":
        version_only = True
    if str(sys.argv[2]) == "system+user":
        user_pkgs = True
    import timeit

    import salt.utils.json

    def run():
        """
        Main run code, when this module is run directly
        """
        pkg_list = WinSoftware(user_pkgs=user_pkgs, version_only=version_only)
        print(
            salt.utils.json.dumps(pkg_list.data, sort_keys=True, indent=4)
        )  # pylint: disable=superfluous-parens
        print("Total: {}".format(len(pkg_list)))  # pylint: disable=superfluous-parens

    print(
        "Time Taken: {}".format(timeit.timeit(run, number=1))
    )  # pylint: disable=superfluous-parens


if __name__ == "__main__":
    __main()



"""
Test ext_nodes master_tops module
"""

import subprocess
import textwrap

import pytest

import salt.tops.ext_nodes as ext_nodes
import salt.utils.stringutils
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        ext_nodes: {
            "__opts__": {
                "master_tops": {
                    # Since ext_nodes runs the command with shell=True,
                    # this will keep "command not found" errors from
                    # showing up on the console. We'll be mocking the
                    # communicate results anyway.
                    "ext_nodes": "echo",
                }
            }
        }
    }


def test_ext_nodes():
    """
    Confirm that subprocess.Popen works as expected and does not raise an
    exception (see https://github.com/saltstack/salt/pull/46863).
    """
    stdout = salt.utils.stringutils.to_bytes(
        textwrap.dedent(
            """\
        classes:
            - one
            - two"""
        )
    )
    run_mock = MagicMock()
    run_mock.return_value.stdout = stdout
    with patch.object(subprocess, "run", run_mock):
        ret = ext_nodes.top(opts={"id": "foo"})
    assert ret == {"base": ["one", "two"]}
    run_mock.assert_called_once_with(["echo", "foo"], check=True, stdout=-1)


def test_ext_nodes_with_environment():
    """
    Same as above, but also tests that the matches are assigned to the proper
    environment if one is returned by the ext_nodes command.
    """
    stdout = salt.utils.stringutils.to_bytes(
        textwrap.dedent(
            """\
        classes:
            - one
            - two
        environment: dev"""
        )
    )
    run_mock = MagicMock()
    run_mock.return_value.stdout = stdout
    with patch.object(subprocess, "run", run_mock):
        ret = ext_nodes.top(opts={"id": "foo"})
    assert ret == {"dev": ["one", "two"]}
    run_mock.assert_called_once_with(["echo", "foo"], check=True, stdout=-1)



"""
Functions to work with MessagePack
"""
import logging

log = logging.getLogger(__name__)

HAS_MSGPACK = False
try:
    import msgpack

    # There is a serialization issue on ARM and potentially other platforms for some msgpack bindings, check for it
    if (
        msgpack.version >= (0, 4, 0)
        and msgpack.loads(msgpack.dumps([1, 2, 3], use_bin_type=False), use_list=True)
        is None
    ):
        raise ImportError
    elif msgpack.loads(msgpack.dumps([1, 2, 3]), use_list=True) is None:
        raise ImportError
    HAS_MSGPACK = True
except ImportError:
    try:
        import msgpack_pure as msgpack  # pylint: disable=import-error

        HAS_MSGPACK = True
    except ImportError:
        pass
        # Don't exit if msgpack is not available, this is to make local mode work without msgpack
        # sys.exit(salt.defaults.exitcodes.EX_GENERIC)

if HAS_MSGPACK and hasattr(msgpack, "exceptions"):
    exceptions = msgpack.exceptions
else:

    class PackValueError(Exception):
        """
        older versions of msgpack do not have PackValueError
        """

    class _exceptions:
        """
        older versions of msgpack do not have an exceptions module
        """

        PackValueError = PackValueError()

    exceptions = _exceptions()

# One-to-one mappings
Packer = msgpack.Packer
ExtType = msgpack.ExtType
version = (0, 0, 0) if not HAS_MSGPACK else msgpack.version


def _sanitize_msgpack_kwargs(kwargs):
    """
    Clean up msgpack keyword arguments based on the version
    https://github.com/msgpack/msgpack-python/blob/master/ChangeLog.rst
    """
    assert isinstance(kwargs, dict)
    if version < (0, 6, 0) and kwargs.pop("strict_map_key", None) is not None:
        log.info("removing unsupported `strict_map_key` argument from msgpack call")
    if version < (0, 5, 2) and kwargs.pop("raw", None) is not None:
        log.info("removing unsupported `raw` argument from msgpack call")
    if version < (0, 4, 0) and kwargs.pop("use_bin_type", None) is not None:
        log.info("removing unsupported `use_bin_type` argument from msgpack call")
    if version >= (1, 0, 0) and kwargs.pop("encoding", None) is not None:
        log.debug("removing unsupported `encoding` argument from msgpack call")

    return kwargs


def _sanitize_msgpack_unpack_kwargs(kwargs):
    """
    Clean up msgpack keyword arguments for unpack operations, based on
    the version
    https://github.com/msgpack/msgpack-python/blob/master/ChangeLog.rst
    """
    assert isinstance(kwargs, dict)
    if version >= (1, 0, 0):
        kwargs.setdefault("raw", True)
        kwargs.setdefault("strict_map_key", False)
    return _sanitize_msgpack_kwargs(kwargs)


def _add_msgpack_unpack_kwargs(kwargs):
    """
    Add any msgpack unpack kwargs here.

    max_buffer_size: will make sure the buffer is set to a minimum
    of 100MiB in versions >=6 and <1.0
    """
    assert isinstance(kwargs, dict)
    if version >= (0, 6, 0) and version < (1, 0, 0):
        kwargs["max_buffer_size"] = 100 * 1024 * 1024
    return _sanitize_msgpack_unpack_kwargs(kwargs)


class Unpacker(msgpack.Unpacker):
    """
    Wraps the msgpack.Unpacker and removes non-relevant arguments
    """

    def __init__(self, *args, **kwargs):
        msgpack.Unpacker.__init__(self, *args, **_add_msgpack_unpack_kwargs(kwargs))


def pack(o, stream, **kwargs):
    """
    .. versionadded:: 2018.3.4

    Wraps msgpack.pack and ensures that the passed object is unwrapped if it is
    a proxy.

    By default, this function uses the msgpack module and falls back to
    msgpack_pure, if the msgpack is not available.
    """
    # Writes to a stream, there is no return
    msgpack.pack(o, stream, **_sanitize_msgpack_kwargs(kwargs))


def packb(o, **kwargs):
    """
    .. versionadded:: 2018.3.4

    Wraps msgpack.packb and ensures that the passed object is unwrapped if it
    is a proxy.

    By default, this function uses the msgpack module and falls back to
    msgpack_pure, if the msgpack is not available.
    """
    return msgpack.packb(o, **_sanitize_msgpack_kwargs(kwargs))


def unpack(stream, **kwargs):
    """
    .. versionadded:: 2018.3.4

    Wraps msgpack.unpack.

    By default, this function uses the msgpack module and falls back to
    msgpack_pure, if the msgpack is not available.
    """
    return msgpack.unpack(stream, **_sanitize_msgpack_unpack_kwargs(kwargs))


def unpackb(packed, **kwargs):
    """
    .. versionadded:: 2018.3.4

    Wraps msgpack.unpack.

    By default, this function uses the msgpack module and falls back to
    msgpack_pure.
    """
    return msgpack.unpackb(packed, **_sanitize_msgpack_unpack_kwargs(kwargs))


# alias for compatibility to simplejson/marshal/pickle.
load = unpack
loads = unpackb

dump = pack
dumps = packb



"""
integration tests for nilirt_ip
"""

import configparser
import re
import shutil
import time

import pytest

import salt.utils.files
import salt.utils.platform

try:
    import pyiface
    from pyiface.ifreqioctls import IFF_LOOPBACK, IFF_RUNNING
except ImportError:
    pyiface = None

try:
    from requests.structures import CaseInsensitiveDict
except ImportError:
    CaseInsensitiveDict = None

INTERFACE_FOR_TEST = "eth1"


pytestmark = [
    pytest.mark.skip_if_not_root,
    pytest.mark.destructive_test,
    pytest.mark.skipif(
        'grains["os_family"] != "NILinuxRT"',
        reason="Tests applicable only to NILinuxRT",
    ),
    pytest.mark.skip_initial_gh_actions_failure(
        reason="This was skipped on older golden images and is failing on newer."
    ),
    pytest.mark.skipif(
        pyiface is None, reason="The python pyiface package is not installed"
    ),
    pytest.mark.skipif(
        CaseInsensitiveDict is None,
        reason="The python package requests is not installed",
    ),
]


def _connected(interface):
    """
    Check if an interface is up or down
    :param interface: pyiface.Interface object
    :return: True, if interface is up, otherwise False.
    """
    return interface.flags & IFF_RUNNING != 0


@pytest.fixture(scope="module")
def interfaces():
    return [
        interface
        for interface in pyiface.getIfaces()
        if interface.flags & IFF_LOOPBACK == 0
    ]


@pytest.fixture(scope="module")
def ethercat_installed(grains):
    """
    Check if ethercat is installed.

    :return: True if ethercat is installed, otherwise False.
    """
    if grains["lsb_distrib_id"] != "nilrt":
        return False

    with salt.utils.files.fopen("/etc/natinst/share/ni-rt.ini", "r") as config_file:
        config_parser = configparser.RawConfigParser(dict_type=CaseInsensitiveDict)
        config_parser.readfp(config_file)
        protocols = config_parser.get(
            "lvrt", "AdditionalNetworkProtocols", fallback=""
        ).lower()
        return "ethercat" in protocols


@pytest.fixture
def ip(modules, grains, interfaces):
    try:
        # save files from var/lib/connman*
        if grains.get("lsb_distrib_id") == "nilrt":
            shutil.move("/etc/natinst/share/ni-rt.ini", "/tmp/ni-rt.ini")
        else:
            shutil.move("/var/lib/connman", "/tmp/connman")

        # Run the test
        yield modules.ip
    finally:
        # Restore files
        if grains.get("lsb_distrib_id") == "nilrt":
            shutil.move("/tmp/ni-rt.ini", "/etc/natinst/share/ni-rt.ini")
            modules.cmd.run("/etc/init.d/networking restart")
        else:
            shutil.move("/tmp/connman", "/var/lib/connman")
            modules.service.restart("connman")
        time.sleep(10)  # wait 10 seconds for connman to be fully loaded
        for interface in interfaces:
            modules.ip.up(interface.name)


def test_down(ip, grains, interfaces):
    """
    Test ip.down function
    """
    for interface in interfaces:
        result = ip.down(interface.name)
        assert result
    info = ip.get_interfaces_details(timeout=300)
    for interface in info["interfaces"]:
        if grains["lsb_distrib_id"] == "nilrt":
            assert interface["adapter_mode"] == "disabled"
        assert not _connected(pyiface.Interface(name=interface["connectionid"]))


def test_up(ip, grains, interfaces):
    """
    Test ip.up function
    """
    # first down all interfaces
    for interface in interfaces:
        ip.down(interface.name)
        assert not _connected(interface)
    # up interfaces
    for interface in interfaces:
        result = ip.up(interface.name)
        assert result
    if grains["lsb_distrib_id"] == "nilrt":
        info = ip.get_interfaces_details(timeout=300)
        for interface in info["interfaces"]:
            assert interface["adapter_mode"] == "tcpip"


def test_set_dhcp_linklocal_all(ip, grains, interfaces):
    """
    Test ip.set_dhcp_linklocal_all function
    """
    for interface in interfaces:
        result = ip.set_dhcp_linklocal_all(interface.name)
        assert result
    info = ip.get_interfaces_details(timeout=300)
    for interface in info["interfaces"]:
        assert interface["ipv4"]["requestmode"] == "dhcp_linklocal"
        if grains["lsb_distrib_id"] == "nilrt":
            assert interface["adapter_mode"] == "tcpip"


def test_set_dhcp_only_all(ip, grains, interfaces):
    """
    Test ip.set_dhcp_only_all function
    """
    if grains["lsb_distrib_id"] != "nilrt":
        raise pytest.skip.Exception(
            "Test not applicable to newer nilrt", _use_item_location=True
        )
    for interface in interfaces:
        result = ip.set_dhcp_only_all(interface.name)
        assert result
    info = ip.get_interfaces_details(timeout=300)
    for interface in info["interfaces"]:
        assert interface["ipv4"]["requestmode"] == "dhcp_only"
        assert interface["adapter_mode"] == "tcpip"


def test_set_linklocal_only_all(ip, grains, interfaces):
    """
    Test ip.set_linklocal_only_all function
    """
    if grains["lsb_distrib_id"] != "nilrt":
        raise pytest.skip.Exception(
            "Test not applicable to newer nilrt", _use_item_location=True
        )
    for interface in interfaces:
        result = ip.set_linklocal_only_all(interface.name)
        assert result
    info = ip.get_interfaces_details(timeout=300)
    for interface in info["interfaces"]:
        assert interface["ipv4"]["requestmode"] == "linklocal_only"
        assert interface["adapter_mode"] == "tcpip"


def test_static_all(ip, grains, interfaces):
    """
    Test ip.set_static_all function
    """
    for interface in interfaces:
        result = ip.set_static_all(
            interface.name,
            "192.168.10.4",
            "255.255.255.0",
            "192.168.10.1",
            "8.8.4.4 8.8.8.8",
        )
        assert result

    info = ip.get_interfaces_details(timeout=300)
    for interface in info["interfaces"]:
        if grains["lsb_distrib_id"] != "nilrt":
            assert "8.8.4.4" in interface["ipv4"]["dns"]
            assert "8.8.8.8" in interface["ipv4"]["dns"]
        else:
            assert interface["ipv4"]["dns"] == ["8.8.4.4"]
            assert interface["adapter_mode"] == "tcpip"
        assert interface["ipv4"]["requestmode"] == "static"
        assert interface["ipv4"]["address"] == "192.168.10.4"
        assert interface["ipv4"]["netmask"] == "255.255.255.0"
        assert interface["ipv4"]["gateway"] == "192.168.10.1"


def test_supported_adapter_modes(ip, grains, ethercat_installed):
    """
    Test supported adapter modes for each interface
    """
    if grains["lsb_distrib_id"] != "nilrt":
        raise pytest.skip.Exception(
            "Test is just for older nilrt distros", _use_item_location=True
        )
    interface_pattern = re.compile("^eth[0-9]+$")
    info = ip.get_interfaces_details(timeout=300)
    for interface in info["interfaces"]:
        if interface["connectionid"] == "eth0":
            assert interface["supported_adapter_modes"] == ["tcpip"]
        else:
            assert "tcpip" in interface["supported_adapter_modes"]
            if not interface_pattern.match(interface["connectionid"]):
                assert "ethercat" not in interface["supported_adapter_modes"]
            elif ethercat_installed:
                assert "ethercat" in interface["supported_adapter_modes"]


def test_ethercat(ip, ethercat_installed):
    """
    Test ip.set_ethercat function
    """
    if not ethercat_installed:
        raise pytest.skip.Exception(
            "Test is just for systems with Ethercat", _use_item_location=True
        )
    assert ip.set_ethercat(INTERFACE_FOR_TEST, 19)
    info = ip.get_interfaces_details(timeout=300)
    for interface in info["interfaces"]:
        if interface["connectionid"] == INTERFACE_FOR_TEST:
            assert interface["adapter_mode"] == "ethercat"
            assert int(interface["ethercat"]["masterid"]) == 19
            break
    assert ip.set_dhcp_linklocal_all(INTERFACE_FOR_TEST)
    info = ip.get_interfaces_details(timeout=300)
    for interface in info["interfaces"]:
        if interface["connectionid"] == INTERFACE_FOR_TEST:
            assert interface["adapter_mode"] == "tcpip"
            assert interface["ipv4"]["requestmode"] == "dhcp_linklocal"
            break


@pytest.mark.destructive_test
def test_dhcp_disable(ip, grains):
    """
    Test cases:
        - dhcp -> disable
        - disable -> dhcp
    """
    if grains["lsb_distrib_id"] == "nilrt":
        raise pytest.skip.Exception(
            "Test is just for newer nilrt distros", _use_item_location=True
        )

    assert ip.set_dhcp_linklocal_all(INTERFACE_FOR_TEST)
    info = ip.get_interfaces_details(timeout=300)
    for interface in info["interfaces"]:
        if interface["connectionid"] == INTERFACE_FOR_TEST:
            assert interface["ipv4"]["requestmode"] == "dhcp_linklocal"
            break

    assert ip.disable(INTERFACE_FOR_TEST)
    info = ip.get_interfaces_details(timeout=300)
    for interface in info["interfaces"]:
        if interface["connectionid"] == INTERFACE_FOR_TEST:
            assert interface["ipv4"]["requestmode"] == "disabled"
            break

    assert ip.set_dhcp_linklocal_all(INTERFACE_FOR_TEST)
    info = ip.get_interfaces_details(timeout=300)
    for interface in info["interfaces"]:
        if interface["connectionid"] == INTERFACE_FOR_TEST:
            assert interface["ipv4"]["requestmode"] == "dhcp_linklocal"
            break


@pytest.mark.destructive_test
def test_dhcp_static(ip, grains):
    """
    Test cases:
        - dhcp -> static
        - static -> dhcp
    """
    if grains["lsb_distrib_id"] == "nilrt":
        raise pytest.skip.Exception(
            "Test is just for newer nilrt distros", _use_item_location=True
        )

    assert ip.set_dhcp_linklocal_all(INTERFACE_FOR_TEST)
    info = ip.get_interfaces_details(timeout=300)
    for interface in info["interfaces"]:
        if interface["connectionid"] == INTERFACE_FOR_TEST:
            assert interface["ipv4"]["requestmode"] == "dhcp_linklocal"
            break

    assert ip.set_static_all(
        INTERFACE_FOR_TEST,
        "192.168.1.125",
        "255.255.255.0",
        "192.168.1.1",
        "8.8.8.8 8.8.8.4",
    )
    info = ip.get_interfaces_details(timeout=300)
    for interface in info["interfaces"]:
        if interface["connectionid"] == INTERFACE_FOR_TEST:
            assert interface["ipv4"]["requestmode"] == "static"
            assert interface["ipv4"]["address"] == "192.168.1.125"
            assert interface["ipv4"]["netmask"] == "255.255.255.0"
            assert "8.8.8.4" in interface["ipv4"]["dns"]
            assert "8.8.8.8" in interface["ipv4"]["dns"]
            break

    assert ip.set_dhcp_linklocal_all(INTERFACE_FOR_TEST)
    info = ip.get_interfaces_details(timeout=300)
    for interface in info["interfaces"]:
        if interface["connectionid"] == INTERFACE_FOR_TEST:
            assert interface["ipv4"]["requestmode"] == "dhcp_linklocal"
            break


@pytest.mark.destructive_test
def test_static_disable(ip, grains):
    """
    Test cases:
        - static -> disable
        - disable -> static
    """
    if grains["lsb_distrib_id"] == "nilrt":
        raise pytest.skip.Exception(
            "Test is just for newer nilrt distros", _use_item_location=True
        )

    assert ip.set_static_all(
        INTERFACE_FOR_TEST,
        "192.168.1.125",
        "255.255.255.0",
        "192.168.1.1",
        "8.8.8.8",
    )
    info = ip.get_interfaces_details(timeout=300)
    for interface in info["interfaces"]:
        if interface["connectionid"] == INTERFACE_FOR_TEST:
            assert interface["ipv4"]["requestmode"] == "static"
            assert interface["ipv4"]["address"] == "192.168.1.125"
            assert interface["ipv4"]["netmask"] == "255.255.255.0"
            assert interface["ipv4"]["dns"] == ["8.8.8.8"]
            break

    assert ip.disable(INTERFACE_FOR_TEST)
    info = ip.get_interfaces_details(timeout=300)
    for interface in info["interfaces"]:
        if interface["connectionid"] == INTERFACE_FOR_TEST:
            assert interface["ipv4"]["requestmode"] == "disabled"
            break

    assert ip.set_static_all(
        INTERFACE_FOR_TEST, "192.168.1.125", "255.255.255.0", "192.168.1.1"
    )
    info = ip.get_interfaces_details(timeout=300)
    for interface in info["interfaces"]:
        if interface["connectionid"] == INTERFACE_FOR_TEST:
            assert interface["ipv4"]["requestmode"] == "static"
            assert interface["ipv4"]["address"] == "192.168.1.125"
            assert interface["ipv4"]["netmask"] == "255.255.255.0"
            assert interface["ipv4"]["dns"] == []
            break



"""
tests.pytests.unit.utils.win_dacl.test_get_sid_string
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Test the get_sid_string function in the win_dacl utility module
"""
import pytest

import salt.utils.win_dacl

# Third-party libs
try:
    import pywintypes

    HAS_WIN32 = True
except ImportError:
    HAS_WIN32 = False

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.skipif(not HAS_WIN32, reason="Requires Win32 libraries"),
]


@pytest.mark.parametrize(
    "principal,expected",
    [
        ("Administrators", "S-1-5-32-544"),  # Normal
        ("adMiniStrAtorS", "S-1-5-32-544"),  # Mixed Case
        ("S-1-5-32-544", "S-1-5-32-544"),  # String SID
        (None, "S-1-0-0"),  # None SID
    ],
)
def test_get_sid_string(principal, expected):
    """
    Validate getting a sid string from a valid pysid object
    """
    sid_obj = salt.utils.win_dacl.get_sid(principal)
    assert isinstance(sid_obj, pywintypes.SIDType)
    assert salt.utils.win_dacl.get_sid_string(sid_obj) == expected



"""
Utilities for working with memcache

:depends:  - python-memcached

This library sets up a connection object for memcache, using the configuration
passed into the get_conn() function. Normally, this is __opts__. Optionally,
a profile or specific host and port may be passed in. If neither profile nor
host and port are provided, the defaults of '`127.0.0.`` and ``11211`` are
used. The following configurations are both valid:

.. code-block:: yaml

    # No profile name
    memcached.host: 127.0.0.1
    memcached.port: 11211

    # One or more profiles defined
    my_memcached_config:
      memcached.host: 127.0.0.1
      memcached.port: 11211

Once configured, the get_conn() function is passed a set of opts, and,
optionally, the name of a profile to be used.

.. code-block:: python

    import salt.utils.memcached_utils.py
    conn = salt.utils.memcached_utils.get_conn(__opts__,
                                              profile='my_memcached_config')

It should be noted that some usages of memcached may require a profile to be
specified, rather than top-level configurations. This being the case, it is
better to always use a named configuration profile, as shown above.
"""


import logging

from salt.exceptions import CommandExecutionError, SaltInvocationError

try:
    import memcache

    HAS_LIBS = True
except ImportError:
    HAS_LIBS = False

DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = 11211
DEFAULT_TIME = 0
DEFAULT_MIN_COMPRESS_LEN = 0

# Set up logging
log = logging.getLogger(__name__)

# Don't shadow built-ins
__func_alias__ = {"set_": "set"}


# Although utils are often directly imported, it is also possible
# to use the loader.
def __virtual__():
    """
    Only load if python-memcached is installed
    """
    return True if HAS_LIBS else False


def get_conn(opts, profile=None, host=None, port=None):
    """
    Return a conn object for accessing memcached
    """
    if not (host and port):
        opts_pillar = opts.get("pillar", {})
        opts_master = opts_pillar.get("master", {})

        opts_merged = {}
        opts_merged.update(opts_master)
        opts_merged.update(opts_pillar)
        opts_merged.update(opts)

        if profile:
            conf = opts_merged.get(profile, {})
        else:
            conf = opts_merged

        host = conf.get("memcached.host", DEFAULT_HOST)
        port = conf.get("memcached.port", DEFAULT_PORT)

    if not str(port).isdigit():
        raise SaltInvocationError("port must be an integer")

    if HAS_LIBS:
        return memcache.Client(["{}:{}".format(host, port)])
    else:
        raise CommandExecutionError(
            "(unable to import memcache, module most likely not installed)"
        )


def _check_stats(conn):
    """
    Helper function to check the stats data passed into it, and raise an
    exception if none are returned. Otherwise, the stats are returned.
    """
    stats = conn.get_stats()
    if not stats:
        raise CommandExecutionError("memcached server is down or does not exist")
    return stats


def set_(
    conn, key, value, time=DEFAULT_TIME, min_compress_len=DEFAULT_MIN_COMPRESS_LEN
):
    """
    Set a key on the memcached server, overwriting the value if it exists.
    """
    if not isinstance(time, int):
        raise SaltInvocationError("'time' must be an integer")
    if not isinstance(min_compress_len, int):
        raise SaltInvocationError("'min_compress_len' must be an integer")
    _check_stats(conn)
    return conn.set(key, value, time, min_compress_len)


def get(conn, key):
    """
    Retrieve value for a key
    """
    _check_stats(conn)
    return conn.get(key)



"""
Integration tests for the zookeeper states
"""

import logging

import pytest
from saltfactories.utils import random_string

pytest.importorskip("kazoo")
pytest.importorskip("docker")

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("dockerd"),
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module")
def minion_config_overrides(zookeeper_port):
    zookeeper_grains = {
        "prod": {
            "hosts": "localhost:{}".format(zookeeper_port),
            "default_acl": [
                {
                    "username": "daniel",
                    "password": "test",
                    "read": True,
                    "write": True,
                    "create": True,
                    "delete": True,
                    "admin": True,
                }
            ],
            "username": "daniel",
            "password": "test",
        },
        "hosts": "localhost:{}".format(zookeeper_port),
        "default_acl": [
            {
                "username": "daniel",
                "password": "test",
                "read": True,
                "write": True,
                "create": True,
                "delete": True,
                "admin": True,
            }
        ],
        "username": "daniel",
        "password": "test",
    }
    return {"grains": {"zookeeper": zookeeper_grains}}


@pytest.fixture(scope="module")
def zookeeper_container(salt_factories):
    container = salt_factories.get_container(
        random_string("zookeeper-"),
        "ghcr.io/saltstack/salt-ci-containers/zookeeper",
        container_run_kwargs={
            "ports": {
                "2181/tcp": None,
            }
        },
        pull_before_start=True,
        skip_on_pull_failure=True,
        skip_if_docker_client_not_connectable=True,
    )
    with container.started() as factory:
        yield factory


@pytest.fixture(scope="module")
def zookeeper_port(zookeeper_container):
    return zookeeper_container.get_host_port_binding(2181, protocol="tcp", ipv6=False)


@pytest.fixture(scope="module")
def zookeeper(zookeeper_container, states):
    return states.zookeeper


def test_zookeeper_present(zookeeper):
    ret = zookeeper.present(name="/test/name-1", value="testuser", makepath=True)
    assert ret.result is True


def test_zookeeper_present_acls_and_profile(zookeeper):
    ret = zookeeper.present(name="/test/name-2", value="testuser", makepath=True)
    assert ret.result is True
    ret = zookeeper.present(
        name="/test/name-2",
        value="daniel",
        acls=[
            {
                "username": "daniel",
                "password": "test",
                "read": True,
                "admin": True,
                "write": True,
            },
            {"username": "testuser", "password": "test", "read": True},
        ],
        profile="prod",
    )
    assert ret.result is True


def test_zookeeper_absent(zookeeper):
    ret = zookeeper.present(name="/test/name-3", value="testuser", makepath=True)
    assert ret.result is True

    ret = zookeeper.absent(name="/test/name-3")
    assert ret.result is True
    assert ret.changes

    ret = zookeeper.absent(name="/test/name-3")
    assert ret.result is True
    assert not ret.changes


def test_zookeeper_acls(zookeeper):
    ret = zookeeper.acls(
        name="/test/name-4",
        acls=[
            {
                "username": "daniel",
                "password": "test",
                "read": True,
                "admin": True,
                "write": True,
            },
            {"username": "testuser", "password": "test", "read": True},
        ],
    )
    assert ret.result is False

    ret = zookeeper.present(name="/test/name-4", value="testuser", makepath=True)
    assert ret.result is True

    ret = zookeeper.acls(
        name="/test/name-4",
        acls=[
            {
                "username": "daniel",
                "password": "test",
                "read": True,
                "admin": True,
                "write": True,
            },
            {"username": "testuser", "password": "test", "read": True},
        ],
    )
    assert ret.result is True



"""
This package contains the runtime hooks support code for when Salt is pacakged with PyInstaller.
"""
import io
import logging
import os
import subprocess
import sys

import salt.utils.vt

log = logging.getLogger(__name__)


def clean_pyinstaller_vars(environ):
    """
    Restore or cleanup PyInstaller specific environent variable behavior.
    """
    if environ is None:
        environ = dict(os.environ)
    # When Salt is bundled with tiamat, it MUST NOT contain LD_LIBRARY_PATH
    # when shelling out, or, at least the value of LD_LIBRARY_PATH set by
    # pyinstaller.
    # See:
    #  https://pyinstaller.readthedocs.io/en/stable/runtime-information.html#ld-library-path-libpath-considerations
    for varname in ("LD_LIBRARY_PATH", "LIBPATH"):
        original_varname = "{}_ORIG".format(varname)
        if varname in environ and environ[varname] == sys._MEIPASS:
            # If we find the varname on the user provided environment we need to at least
            # check if it's not the value set by PyInstaller, if it is, remove it.
            log.debug(
                "User provided environment variable %r with value %r which is "
                "the value that PyInstaller set's. Removing it",
                varname,
                environ[varname],
            )
            environ.pop(varname)

        if original_varname in environ and varname not in environ:
            # We found the original variable set by PyInstaller, and we didn't find
            # any user provided variable, let's rename it.
            log.debug(
                "The %r variable was found in the passed environment, renaming it to %r",
                original_varname,
                varname,
            )
            environ[varname] = environ.pop(original_varname)

        if varname not in environ:
            if original_varname in os.environ:
                log.debug(
                    "Renaming environment variable %r to %r", original_varname, varname
                )
                environ[varname] = os.environ[original_varname]
            elif varname in os.environ:
                # Override the system environ variable with an empty one
                log.debug("Setting environment variable %r to an empty string", varname)
                environ[varname] = ""
    return environ


class PyinstallerPopen(subprocess.Popen):
    def __init__(self, *args, **kwargs):
        kwargs["env"] = clean_pyinstaller_vars(kwargs.pop("env", None))
        super().__init__(*args, **kwargs)

    # From https://github.com/pyinstaller/pyinstaller/blob/v5.1/PyInstaller/hooks/rthooks/pyi_rth_subprocess.py
    #
    # In windowed mode, force any unused pipes (stdin, stdout and stderr) to be DEVNULL instead of inheriting the
    # invalid corresponding handles from this parent process.
    if sys.platform == "win32" and not isinstance(sys.stdout, io.IOBase):

        def _get_handles(self, stdin, stdout, stderr):
            stdin, stdout, stderr = (
                subprocess.DEVNULL if pipe is None else pipe
                for pipe in (stdin, stdout, stderr)
            )
            return super()._get_handles(stdin, stdout, stderr)


class PyinstallerTerminal(salt.utils.vt.Terminal):  # pylint: disable=abstract-method
    def __init__(self, *args, **kwargs):
        kwargs["env"] = clean_pyinstaller_vars(kwargs.pop("env", None))
        super().__init__(*args, **kwargs)



import pickle

from salt.defaults import _Constant


def test_pickle_constants():
    """
    That that we can pickle and unpickle constants.
    """
    constant = _Constant("Foo", 123)
    sdata = pickle.dumps(constant)
    odata = pickle.loads(sdata)
    assert odata == constant



#!/usr/bin/env python
"""
The setup script for sodium_grabber
"""

# pylint: disable=C0111,E1101,E1103,F0401,W0611

from distutils.core import Extension, setup
from os import path

HERE = path.dirname(__file__)

SETUP_KWARGS = {}
sodium_grabber = Extension(
    "sodium_grabber",
    sources=[path.join(HERE, "sodium_grabber.c")],
    libraries=["sodium"],
)
SETUP_KWARGS["ext_modules"] = [sodium_grabber]
SETUP_KWARGS["name"] = "sodium_grabber"

if __name__ == "__main__":
    setup(**SETUP_KWARGS)



#
# Copyright 2015 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os

from salt.modules.inspectlib.dbhandle import DBHandle
from salt.modules.inspectlib.exceptions import InspectorSnapshotException


class EnvLoader:
    """
    Load environment.
    """

    PID_FILE = "_minion_collector.pid"
    DB_FILE = "_minion_collector.db"
    DEFAULT_PID_PATH = "/var/run"
    DEFAULT_CACHE_PATH = "/var/cache/salt"

    def __init__(self, cachedir=None, piddir=None, pidfilename=None):
        """
        Constructor.

        :param options:
        :param db_path:
        :param pid_file:
        """
        if not cachedir and "__salt__" in globals():
            cachedir = globals().get("__salt__")["config.get"]("inspector.db", "")

        self.dbfile = os.path.join(cachedir or self.DEFAULT_CACHE_PATH, self.DB_FILE)
        self.db = DBHandle(self.dbfile)

        if not piddir and "__salt__" in globals():
            piddir = globals().get("__salt__")["config.get"]("inspector.pid", "")
        self.pidfile = os.path.join(
            piddir or self.DEFAULT_PID_PATH, pidfilename or self.PID_FILE
        )



"""
Tests for win_path states
"""

import copy

import pytest

import salt.states.win_path as win_path
from tests.support.mock import MagicMock, Mock, patch


@pytest.fixture
def name():
    return "salt"


@pytest.fixture
def configure_loader_modules():
    return {win_path: {}}


def test_absent(name):
    """
    Test various cases for win_path.absent
    """
    ret_base = {"name": name, "result": True, "changes": {}}

    def _mock(retval):
        # Return a new MagicMock for each test case
        return MagicMock(side_effect=retval)

    # We don't really want to run the remove func
    with patch.dict(win_path.__salt__, {"win_path.remove": Mock()}):

        # Test mode OFF
        with patch.dict(win_path.__opts__, {"test": False}):

            # Test already absent
            with patch.dict(win_path.__salt__, {"win_path.exists": _mock([False])}):
                ret = copy.deepcopy(ret_base)
                ret["comment"] = "{} is not in the PATH".format(name)
                ret["result"] = True
                assert win_path.absent(name) == ret

            # Test successful removal
            with patch.dict(
                win_path.__salt__, {"win_path.exists": _mock([True, False])}
            ):
                ret = copy.deepcopy(ret_base)
                ret["comment"] = "Removed {} from the PATH".format(name)
                ret["changes"]["removed"] = name
                ret["result"] = True
                assert win_path.absent(name) == ret

            # Test unsucessful removal
            with patch.dict(
                win_path.__salt__, {"win_path.exists": _mock([True, True])}
            ):
                ret = copy.deepcopy(ret_base)
                ret["comment"] = "Failed to remove {} from the PATH".format(name)
                ret["result"] = False
                assert win_path.absent(name) == ret

        # Test mode ON
        with patch.dict(win_path.__opts__, {"test": True}):

            # Test already absent
            with patch.dict(win_path.__salt__, {"win_path.exists": _mock([False])}):
                ret = copy.deepcopy(ret_base)
                ret["comment"] = "{} is not in the PATH".format(name)
                ret["result"] = True
                assert win_path.absent(name) == ret

            # Test the test-mode return
            with patch.dict(win_path.__salt__, {"win_path.exists": _mock([True])}):
                ret = copy.deepcopy(ret_base)
                ret["comment"] = "{} would be removed from the PATH".format(name)
                ret["result"] = None
                assert win_path.absent(name) == ret


def test_exists_invalid_index(name):
    """
    Tests win_path.exists when a non-integer index is specified.
    """
    ret = win_path.exists(name, index="foo")
    assert ret == {
        "name": name,
        "changes": {},
        "result": False,
        "comment": "Index must be an integer",
    }


def test_exists_add_no_index_success(name):
    """
    Tests win_path.exists when the directory isn't already in the PATH and
    no index is specified (successful run).
    """
    add_mock = MagicMock(return_value=True)
    rehash_mock = MagicMock(return_value=True)
    dunder_salt = {
        "win_path.get_path": MagicMock(
            side_effect=[["foo", "bar", "baz"], ["foo", "bar", "baz", name]]
        ),
        "win_path.add": add_mock,
        "win_path.rehash": rehash_mock,
    }
    dunder_opts = {"test": False}

    with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
        win_path.__opts__, dunder_opts
    ):
        ret = win_path.exists(name)

    add_mock.assert_called_once_with(name, index=None, rehash=False)
    rehash_mock.assert_called_once_with()
    assert ret == {
        "name": name,
        "changes": {"index": {"old": None, "new": 3}},
        "result": True,
        "comment": "Added {} to the PATH.".format(name),
    }


def test_exists_add_no_index_failure(name):
    """
    Tests win_path.exists when the directory isn't already in the PATH and
    no index is specified (failed run).
    """
    add_mock = MagicMock(return_value=False)
    rehash_mock = MagicMock(return_value=True)
    dunder_salt = {
        "win_path.get_path": MagicMock(
            side_effect=[["foo", "bar", "baz"], ["foo", "bar", "baz"]]
        ),
        "win_path.add": add_mock,
        "win_path.rehash": rehash_mock,
    }
    dunder_opts = {"test": False}

    with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
        win_path.__opts__, dunder_opts
    ):
        ret = win_path.exists(name)

    add_mock.assert_called_once_with(name, index=None, rehash=False)
    rehash_mock.assert_not_called()
    assert ret == {
        "name": name,
        "changes": {},
        "result": False,
        "comment": "Failed to add {} to the PATH.".format(name),
    }


def test_exists_add_no_index_failure_exception(name):
    """
    Tests win_path.exists when the directory isn't already in the PATH and
    no index is specified (failed run due to exception).
    """
    add_mock = MagicMock(side_effect=Exception("Global Thermonuclear War"))
    rehash_mock = MagicMock(return_value=True)
    dunder_salt = {
        "win_path.get_path": MagicMock(
            side_effect=[["foo", "bar", "baz"], ["foo", "bar", "baz"]]
        ),
        "win_path.add": add_mock,
        "win_path.rehash": rehash_mock,
    }
    dunder_opts = {"test": False}

    with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
        win_path.__opts__, dunder_opts
    ):
        ret = win_path.exists(name)

    add_mock.assert_called_once_with(name, index=None, rehash=False)
    rehash_mock.assert_not_called()
    assert ret == {
        "name": name,
        "changes": {},
        "result": False,
        "comment": (
            "Encountered error: Global Thermonuclear War. "
            "Failed to add {} to the PATH.".format(name)
        ),
    }


def test_exists_change_index_success(name):
    """
    Tests win_path.exists when the directory is already in the PATH and
    needs to be moved to a different position (successful run).
    """
    add_mock = MagicMock(return_value=True)
    rehash_mock = MagicMock(return_value=True)
    dunder_salt = {
        "win_path.get_path": MagicMock(
            side_effect=[["foo", "bar", "baz", name], [name, "foo", "bar", "baz"]]
        ),
        "win_path.add": add_mock,
        "win_path.rehash": rehash_mock,
    }
    dunder_opts = {"test": False}

    with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
        win_path.__opts__, dunder_opts
    ):
        ret = win_path.exists(name, index=0)

    add_mock.assert_called_once_with(name, index=0, rehash=False)
    rehash_mock.assert_called_once()
    assert ret == {
        "name": name,
        "changes": {"index": {"old": 3, "new": 0}},
        "result": True,
        "comment": "Moved {} from index 3 to 0.".format(name),
    }


def test_exists_change_negative_index_success(name):
    """
    Tests win_path.exists when the directory is already in the PATH and
    needs to be moved to a different position (successful run).

    This tests a negative index.
    """
    add_mock = MagicMock(return_value=True)
    rehash_mock = MagicMock(return_value=True)
    dunder_salt = {
        "win_path.get_path": MagicMock(
            side_effect=[["foo", "bar", name, "baz"], ["foo", "bar", "baz", name]]
        ),
        "win_path.add": add_mock,
        "win_path.rehash": rehash_mock,
    }
    dunder_opts = {"test": False}

    with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
        win_path.__opts__, dunder_opts
    ):
        ret = win_path.exists(name, index=-1)

    add_mock.assert_called_once_with(name, index=-1, rehash=False)
    rehash_mock.assert_called_once()
    assert ret == {
        "name": name,
        "changes": {"index": {"old": -2, "new": -1}},
        "result": True,
        "comment": "Moved {} from index -2 to -1.".format(name),
    }


def test_exists_change_index_add_exception(name):
    """
    Tests win_path.exists when the directory is already in the PATH but an
    exception is raised when we attempt to add the key to its new location.
    """
    add_mock = MagicMock(side_effect=Exception("Global Thermonuclear War"))
    rehash_mock = MagicMock(return_value=True)
    dunder_salt = {
        "win_path.get_path": MagicMock(
            side_effect=[["foo", "bar", "baz", name], ["foo", "bar", "baz", name]]
        ),
        "win_path.add": add_mock,
        "win_path.rehash": rehash_mock,
    }
    dunder_opts = {"test": False}

    with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
        win_path.__opts__, dunder_opts
    ):
        ret = win_path.exists(name, index=0)

    add_mock.assert_called_once_with(name, index=0, rehash=False)
    rehash_mock.assert_not_called()
    assert ret == {
        "name": name,
        "changes": {},
        "result": False,
        "comment": (
            "Encountered error: Global Thermonuclear War. "
            "Failed to move {} from index 3 to 0.".format(name)
        ),
    }


def test_exists_change_negative_index_add_exception(name):
    """
    Tests win_path.exists when the directory is already in the PATH but an
    exception is raised when we attempt to add the key to its new location.

    This tests a negative index.
    """
    add_mock = MagicMock(side_effect=Exception("Global Thermonuclear War"))
    rehash_mock = MagicMock(return_value=True)
    dunder_salt = {
        "win_path.get_path": MagicMock(
            side_effect=[["foo", "bar", name, "baz"], ["foo", "bar", name, "baz"]]
        ),
        "win_path.add": add_mock,
        "win_path.rehash": rehash_mock,
    }
    dunder_opts = {"test": False}

    with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
        win_path.__opts__, dunder_opts
    ):
        ret = win_path.exists(name, index=-1)

    add_mock.assert_called_once_with(name, index=-1, rehash=False)
    rehash_mock.assert_not_called()
    assert ret == {
        "name": name,
        "changes": {},
        "result": False,
        "comment": (
            "Encountered error: Global Thermonuclear War. "
            "Failed to move {} from index -2 to -1.".format(name)
        ),
    }


def test_exists_change_index_failure(name):
    """
    Tests win_path.exists when the directory is already in the PATH and
    needs to be moved to a different position (failed run).
    """
    add_mock = MagicMock(return_value=False)
    rehash_mock = MagicMock(return_value=True)
    dunder_salt = {
        "win_path.get_path": MagicMock(
            side_effect=[["foo", "bar", "baz", name], ["foo", "bar", "baz", name]]
        ),
        "win_path.add": add_mock,
        "win_path.rehash": rehash_mock,
    }
    dunder_opts = {"test": False}

    with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
        win_path.__opts__, dunder_opts
    ):
        ret = win_path.exists(name, index=0)

    add_mock.assert_called_once_with(name, index=0, rehash=False)
    rehash_mock.assert_not_called()
    assert ret == {
        "name": name,
        "changes": {},
        "result": False,
        "comment": "Failed to move {} from index 3 to 0.".format(name),
    }


def test_exists_change_negative_index_failure(name):
    """
    Tests win_path.exists when the directory is already in the PATH and
    needs to be moved to a different position (failed run).

    This tests a negative index.
    """
    add_mock = MagicMock(return_value=False)
    rehash_mock = MagicMock(return_value=True)
    dunder_salt = {
        "win_path.get_path": MagicMock(
            side_effect=[["foo", "bar", name, "baz"], ["foo", "bar", name, "baz"]]
        ),
        "win_path.add": add_mock,
        "win_path.rehash": rehash_mock,
    }
    dunder_opts = {"test": False}

    with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
        win_path.__opts__, dunder_opts
    ):
        ret = win_path.exists(name, index=-1)

    add_mock.assert_called_once_with(name, index=-1, rehash=False)
    rehash_mock.assert_not_called()
    assert ret == {
        "name": name,
        "changes": {},
        "result": False,
        "comment": "Failed to move {} from index -2 to -1.".format(name),
    }


def test_exists_change_index_test_mode(name):
    """
    Tests win_path.exists when the directory is already in the PATH and
    needs to be moved to a different position (test mode enabled).
    """
    add_mock = Mock()
    rehash_mock = MagicMock(return_value=True)
    dunder_salt = {
        "win_path.get_path": MagicMock(side_effect=[["foo", "bar", "baz", name]]),
        "win_path.add": add_mock,
        "win_path.rehash": rehash_mock,
    }
    dunder_opts = {"test": True}

    with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
        win_path.__opts__, dunder_opts
    ):
        ret = win_path.exists(name, index=0)

    add_mock.assert_not_called()
    rehash_mock.assert_not_called()
    assert ret == {
        "name": name,
        "changes": {"index": {"old": 3, "new": 0}},
        "result": None,
        "comment": "{} would be moved from index 3 to 0.".format(name),
    }


def test_exists_change_negative_index_test_mode(name):
    """
    Tests win_path.exists when the directory is already in the PATH and
    needs to be moved to a different position (test mode enabled).
    """
    add_mock = Mock()
    rehash_mock = MagicMock(return_value=True)
    dunder_salt = {
        "win_path.get_path": MagicMock(side_effect=[["foo", "bar", name, "baz"]]),
        "win_path.add": add_mock,
        "win_path.rehash": rehash_mock,
    }
    dunder_opts = {"test": True}

    with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
        win_path.__opts__, dunder_opts
    ):
        ret = win_path.exists(name, index=-1)

    add_mock.assert_not_called()
    rehash_mock.assert_not_called()
    assert ret == {
        "name": name,
        "changes": {"index": {"old": -2, "new": -1}},
        "result": None,
        "comment": "{} would be moved from index -2 to -1.".format(name),
    }


def _test_exists_add_already_present(index, test_mode, name):
    """
    Tests win_path.exists when the directory already exists in the PATH.
    Helper function to test both with and without and index, and with test
    mode both disabled and enabled.
    """
    current_path = ["foo", "bar", "baz"]
    if index is None:
        current_path.append(name)
    else:
        pos = index if index >= 0 else len(current_path) + index + 1
        current_path.insert(pos, name)

    add_mock = Mock()
    rehash_mock = MagicMock(return_value=True)
    dunder_salt = {
        "win_path.get_path": MagicMock(side_effect=[current_path]),
        "win_path.add": add_mock,
        "win_path.rehash": rehash_mock,
    }
    dunder_opts = {"test": test_mode}

    with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
        win_path.__opts__, dunder_opts
    ):
        ret = win_path.exists(name, index=index)

    add_mock.assert_not_called()
    rehash_mock.assert_not_called()
    assert ret == {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "{} already exists in the PATH{}.".format(
            name, " at index {}".format(index) if index is not None else ""
        ),
    }


def test_exists_add_no_index_already_present(name):
    _test_exists_add_already_present(None, False, name)


def test_exists_add_no_index_already_present_test_mode(name):
    _test_exists_add_already_present(None, True, name)


def test_exists_add_index_already_present(name):
    _test_exists_add_already_present(1, False, name)
    _test_exists_add_already_present(2, False, name)
    _test_exists_add_already_present(-1, False, name)
    _test_exists_add_already_present(-2, False, name)


def test_exists_add_index_already_present_test_mode(name):
    _test_exists_add_already_present(1, True, name)
    _test_exists_add_already_present(2, True, name)
    _test_exists_add_already_present(-1, True, name)
    _test_exists_add_already_present(-2, True, name)



"""
Apache Libcloud Storage Management
==================================

Connection module for Apache Libcloud Storage (object/blob) management for a full list
of supported clouds, see http://libcloud.readthedocs.io/en/latest/storage/supported_providers.html

Clouds include Amazon S3, Google Storage, Aliyun, Azure Blobs, Ceph, OpenStack swift

.. versionadded:: 2018.3.0

:configuration:
    This module uses a configuration profile for one or multiple Storage providers

    .. code-block:: yaml

        libcloud_storage:
            profile_test1:
              driver: google_storage
              key: GOOG0123456789ABCXYZ
              secret: mysecret
            profile_test2:
              driver: s3
              key: 12345
              secret: mysecret

:depends: apache-libcloud
"""
# keep lint from choking on _get_conn and _cache_id
# pylint: disable=E0602


import logging

import salt.utils.args
import salt.utils.compat
from salt.utils.versions import Version

log = logging.getLogger(__name__)

REQUIRED_LIBCLOUD_VERSION = "1.5.0"
try:
    # pylint: disable=unused-import
    import libcloud
    from libcloud.storage.providers import get_driver

    # pylint: enable=unused-import
    if hasattr(libcloud, "__version__") and Version(libcloud.__version__) < Version(
        REQUIRED_LIBCLOUD_VERSION
    ):
        raise ImportError()
    logging.getLogger("libcloud").setLevel(logging.CRITICAL)
    HAS_LIBCLOUD = True
except ImportError:
    HAS_LIBCLOUD = False


def __virtual__():
    """
    Only load if libcloud libraries exist.
    """
    if not HAS_LIBCLOUD:
        return (
            False,
            "A apache-libcloud library with version at least {} was not found".format(
                REQUIRED_LIBCLOUD_VERSION
            ),
        )
    return True


def _get_driver(profile):
    config = __salt__["config.option"]("libcloud_storage")[profile]
    cls = get_driver(config["driver"])
    args = config.copy()
    del args["driver"]
    args["key"] = config.get("key")
    args["secret"] = config.get("secret", None)
    args["secure"] = config.get("secure", True)
    args["host"] = config.get("host", None)
    args["port"] = config.get("port", None)
    return cls(**args)


def list_containers(profile, **libcloud_kwargs):
    """
    Return a list of containers.

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's list_containers method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_storage.list_containers profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    containers = conn.list_containers(**libcloud_kwargs)
    ret = []
    for container in containers:
        ret.append({"name": container.name, "extra": container.extra})
    return ret


def list_container_objects(container_name, profile, **libcloud_kwargs):
    """
    List container objects (e.g. files) for the given container_id on the given profile

    :param container_name: Container name
    :type  container_name: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's list_container_objects method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_storage.list_container_objects MyFolder profile1
    """
    conn = _get_driver(profile=profile)
    container = conn.get_container(container_name)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    objects = conn.list_container_objects(container, **libcloud_kwargs)
    ret = []
    for obj in objects:
        ret.append(
            {
                "name": obj.name,
                "size": obj.size,
                "hash": obj.hash,
                "container": obj.container.name,
                "extra": obj.extra,
                "meta_data": obj.meta_data,
            }
        )
    return ret


def create_container(container_name, profile, **libcloud_kwargs):
    """
    Create a container in the cloud

    :param container_name: Container name
    :type  container_name: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's create_container method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_storage.create_container MyFolder profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    container = conn.create_container(container_name, **libcloud_kwargs)
    return {"name": container.name, "extra": container.extra}


def get_container(container_name, profile, **libcloud_kwargs):
    """
    List container details for the given container_name on the given profile

    :param container_name: Container name
    :type  container_name: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's get_container method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_storage.get_container MyFolder profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    container = conn.get_container(container_name, **libcloud_kwargs)
    return {"name": container.name, "extra": container.extra}


def get_container_object(container_name, object_name, profile, **libcloud_kwargs):
    """
    Get the details for a container object (file or object in the cloud)

    :param container_name: Container name
    :type  container_name: ``str``

    :param object_name: Object name
    :type  object_name: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's get_container_object method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_storage.get_container_object MyFolder MyFile.xyz profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    obj = conn.get_container_object(container_name, object_name, **libcloud_kwargs)
    return {
        "name": obj.name,
        "size": obj.size,
        "hash": obj.hash,
        "container": obj.container.name,
        "extra": obj.extra,
        "meta_data": obj.meta_data,
    }


def download_object(
    container_name,
    object_name,
    destination_path,
    profile,
    overwrite_existing=False,
    delete_on_failure=True,
    **libcloud_kwargs
):
    """
    Download an object to the specified destination path.

    :param container_name: Container name
    :type  container_name: ``str``

    :param object_name: Object name
    :type  object_name: ``str``

    :param destination_path: Full path to a file or a directory where the
                                incoming file will be saved.
    :type destination_path: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param overwrite_existing: True to overwrite an existing file,
                                defaults to False.
    :type overwrite_existing: ``bool``

    :param delete_on_failure: True to delete a partially downloaded file if
                                the download was not successful (hash
                                mismatch / file size).
    :type delete_on_failure: ``bool``

    :param libcloud_kwargs: Extra arguments for the driver's download_object method
    :type  libcloud_kwargs: ``dict``

    :return: True if an object has been successfully downloaded, False
                otherwise.
    :rtype: ``bool``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_storage.download_object MyFolder me.jpg /tmp/me.jpg profile1

    """
    conn = _get_driver(profile=profile)
    obj = conn.get_object(container_name, object_name)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    return conn.download_object(
        obj, destination_path, overwrite_existing, delete_on_failure, **libcloud_kwargs
    )


def upload_object(
    file_path,
    container_name,
    object_name,
    profile,
    extra=None,
    verify_hash=True,
    headers=None,
    **libcloud_kwargs
):
    """
    Upload an object currently located on a disk.

    :param file_path: Path to the object on disk.
    :type file_path: ``str``

    :param container_name: Destination container.
    :type container_name: ``str``

    :param object_name: Object name.
    :type object_name: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param verify_hash: Verify hash
    :type verify_hash: ``bool``

    :param extra: Extra attributes (driver specific). (optional)
    :type extra: ``dict``

    :param headers: (optional) Additional request headers,
        such as CORS headers. For example:
        headers = {'Access-Control-Allow-Origin': 'http://mozilla.com'}
    :type headers: ``dict``

    :param libcloud_kwargs: Extra arguments for the driver's upload_object method
    :type  libcloud_kwargs: ``dict``

    :return: The object name in the cloud
    :rtype: ``str``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_storage.upload_object /file/to/me.jpg MyFolder me.jpg profile1

    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    container = conn.get_container(container_name)
    obj = conn.upload_object(
        file_path,
        container,
        object_name,
        extra,
        verify_hash,
        headers,
        **libcloud_kwargs
    )
    return obj.name


def delete_object(container_name, object_name, profile, **libcloud_kwargs):
    """
    Delete an object in the cloud

    :param container_name: Container name
    :type  container_name: ``str``

    :param object_name: Object name
    :type  object_name: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's delete_object method
    :type  libcloud_kwargs: ``dict``

    :return: True if an object has been successfully deleted, False
                otherwise.
    :rtype: ``bool``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_storage.delete_object MyFolder me.jpg profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    obj = conn.get_object(container_name, object_name, **libcloud_kwargs)
    return conn.delete_object(obj)


def delete_container(container_name, profile, **libcloud_kwargs):
    """
    Delete an object container in the cloud

    :param container_name: Container name
    :type  container_name: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's delete_container method
    :type  libcloud_kwargs: ``dict``

    :return: True if an object container has been successfully deleted, False
                otherwise.
    :rtype: ``bool``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_storage.delete_container MyFolder profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    container = conn.get_container(container_name)
    return conn.delete_container(container, **libcloud_kwargs)


def extra(method, profile, **libcloud_kwargs):
    """
    Call an extended method on the driver

    :param method: Driver's method name
    :type  method: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's delete_container method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_storage.extra ex_get_permissions google container_name=my_container object_name=me.jpg --out=yaml
    """
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    conn = _get_driver(profile=profile)
    connection_method = getattr(conn, method)
    return connection_method(**libcloud_kwargs)



"""
External Nodes Classifier
=========================

The External Nodes Classifier is a master tops subsystem that retrieves mapping
information from major configuration management systems. One of the most common
external nodes classifiers system is provided by Cobbler and is called
``cobbler-ext-nodes``.

The cobbler-ext-nodes command can be used with this configuration:

.. code-block:: yaml

    master_tops:
      ext_nodes: cobbler-ext-nodes

It is noteworthy that the Salt system does not directly ingest the data
sent from the ``cobbler-ext-nodes`` command, but converts the data into
information that is used by a Salt top file.

Any command can replace the call to 'cobbler-ext-nodes' above, but currently the
data must be formatted in the same way that the standard 'cobbler-ext-nodes'
does.

See (admittedly degenerate and probably not complete) example:

.. code-block:: yaml

    classes:
      - basepackages
      - database

The above essentially is the same as a top.sls containing the following:

.. code-block:: yaml

    base:
      '*':
        - basepackages
        - database

    base:
      '*':
        - basepackages
        - database
"""
import logging
import shlex
import subprocess

import salt.utils.platform
import salt.utils.yaml

if salt.utils.platform.is_windows():
    from salt.utils.win_functions import escape_argument as _cmd_quote
else:
    _cmd_quote = shlex.quote

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only run if properly configured
    """
    if __opts__["master_tops"].get("ext_nodes"):
        return True
    return False


def top(**kwargs):
    """
    Run the command configured
    """
    if "id" not in kwargs["opts"]:
        return {}
    proc = subprocess.run(
        [
            _cmd_quote(part)
            for part in shlex.split(
                __opts__["master_tops"]["ext_nodes"],
                posix=salt.utils.platform.is_windows() is False,
            )
            + [_cmd_quote(kwargs["opts"]["id"])]
        ],
        stdout=subprocess.PIPE,
        check=True,
    )
    ndata = salt.utils.yaml.safe_load(proc.stdout)
    if not ndata:
        log.info("master_tops ext_nodes call did not return any data")
    ret = {}
    if "environment" in ndata:
        env = ndata["environment"]
    else:
        env = "base"

    if "classes" in ndata:
        if isinstance(ndata["classes"], dict):
            ret[env] = list(ndata["classes"])
        elif isinstance(ndata["classes"], list):
            ret[env] = ndata["classes"]
        else:
            return ret
    else:
        log.info(
            'master_tops ext_nodes call did not have a dictionary with a "classes" key.'
        )

    return ret



import enum
import time

import pytest
import requests
import requests.exceptions
from saltfactories.utils import random_string

from salt.utils.etcd_util import HAS_ETCD_V2, HAS_ETCD_V3


class EtcdVersion(enum.Enum):
    v2 = "etcd-v2"
    v3 = "etcd-v3"
    v3_v2_mode = "etcd-v3(v2-mode)"


def etcd_version_ids(enum_value):
    return enum_value.value


@pytest.fixture(scope="module", params=tuple(EtcdVersion), ids=etcd_version_ids)
def etcd_version(request):
    if request.param == EtcdVersion.v2 and not HAS_ETCD_V2:
        pytest.skip("No etcd library installed")
    if request.param != EtcdVersion.v2 and not HAS_ETCD_V3:
        pytest.skip("No etcd3 library installed")
    return request.param


@pytest.fixture(scope="module")
def etcd_container_image_name(etcd_version):
    if etcd_version == EtcdVersion.v2:
        return "ghcr.io/saltstack/salt-ci-containers/etcd:2"
    return "ghcr.io/saltstack/salt-ci-containers/etcd:3"


@pytest.fixture(scope="module")
def etcd_container_name(etcd_version):
    if etcd_version == EtcdVersion.v2:
        return random_string("etcd-v2-server-")
    if etcd_version == EtcdVersion.v3:
        return random_string("etcd-v3-server-")
    return random_string("etcd-v3-in-v2-mode-server-")


@pytest.fixture(scope="module")
def etcd_static_port():
    """
    We return ``None`` because we want docker to assign the host port for us.

    Return a port number to override the above behavior.
    """
    return None


def confirm_container_started(timeout_at, container):
    etcd_port = container.get_host_port_binding(2379, protocol="tcp", ipv6=False)
    sleeptime = 1
    while time.time() <= timeout_at:
        try:
            response = requests.get("http://localhost:{}/version".format(etcd_port))
            try:
                version = response.json()
                if "etcdserver" in version:
                    # etcd >= v3
                    break
            except ValueError:
                # etcd v2
                if "etcd 2." in response.text:
                    break
        except requests.exceptions.ConnectionError:
            pass
        time.sleep(sleeptime)
        sleeptime *= 2
    else:
        return False
    return True


@pytest.fixture(scope="module")
def etcd_container(
    salt_factories,
    etcd_container_name,
    etcd_container_image_name,
    etcd_static_port,
    etcd_version,
):
    container_environment = {
        "ALLOW_NONE_AUTHENTICATION": "yes",
    }
    if etcd_version == EtcdVersion.v3_v2_mode:
        container_environment["ETCD_ENABLE_V2"] = "true"
    container = salt_factories.get_container(
        etcd_container_name,
        image_name=etcd_container_image_name,
        container_run_kwargs={
            "environment": container_environment,
            "ports": {"2379/tcp": etcd_static_port},
        },
        pull_before_start=True,
        skip_on_pull_failure=True,
        skip_if_docker_client_not_connectable=True,
    )
    container.container_start_check(confirm_container_started, container)
    with container.started() as factory:
        yield factory


@pytest.fixture(scope="module")
def etcd_port(etcd_container):
    return etcd_container.get_host_port_binding(2379, protocol="tcp", ipv6=False)


@pytest.fixture(scope="module")
def profile_name():
    return "etcd_util_profile"


@pytest.fixture(scope="module")
def etcd_profile(profile_name, etcd_port, etcd_version):
    profile = {
        profile_name: {
            "etcd.host": "127.0.0.1",
            "etcd.port": etcd_port,
            "etcd.require_v2": etcd_version in (EtcdVersion.v2, EtcdVersion.v3_v2_mode),
        }
    }

    return profile



"""
Unit tests for salt.utils.templates.py
"""
import logging
from pathlib import PurePath, PurePosixPath

import pytest

from salt.utils.templates import wrap_tmpl_func, generate_sls_context
from tests.support.mock import patch


log = logging.getLogger(__name__)


class MockRender:
    def __call__(self, tplstr, context, tmplpath=None):
        self.tplstr = tplstr
        self.context = context
        self.tmplpath = tmplpath
        return tplstr


def _test_generated_sls_context(tmplpath, sls, **expected):
    """Generic SLS Context Test"""
    # DeNormalize tmplpath
    tmplpath = str(PurePath(PurePosixPath(tmplpath)))
    if tmplpath.startswith("\\"):
        tmplpath = "C:{}".format(tmplpath)
    expected["tplpath"] = tmplpath
    actual = generate_sls_context(tmplpath, sls)
    assert {key: actual[key] for key in expected if key in actual} == actual


def test_sls_context_call(tmp_path):
    """Check that generate_sls_context is called with proper parameters"""
    sls = "foo.bar"
    slsfile = tmp_path / "foo" / "bar.sls"
    slsfile.parent.mkdir()
    slsfile.write_text("{{ slspath }}")
    context = {"opts": {}, "saltenv": "base", "sls": sls}
    render = MockRender()
    with patch("salt.utils.templates.generate_sls_context") as generate_sls_context:
        wrapped = wrap_tmpl_func(render)
        res = wrapped(str(slsfile), context=context, tmplpath=str(slsfile))
        generate_sls_context.assert_called_with(str(slsfile), sls)


def test_sls_context_no_call(tmp_path):
    """Check that generate_sls_context is not called if sls is not set"""
    sls = "foo.bar"
    slsfile = tmp_path / "foo" / "bar.sls"
    slsfile.parent.mkdir()
    slsfile.write_text("{{ slspath }}")
    context = {"opts": {}, "saltenv": "base"}
    render = MockRender()
    with patch("salt.utils.templates.generate_sls_context") as generate_sls_context:
        wrapped = wrap_tmpl_func(render)
        res = wrapped(str(slsfile), context=context, tmplpath=str(slsfile))
        generate_sls_context.assert_not_called()


def test_generate_sls_context__top_level():
    """generate_sls_context - top_level Use case"""
    _test_generated_sls_context(
        "/tmp/boo.sls",
        "boo",
        tplfile="boo.sls",
        tpldir=".",
        tpldot="",
        slsdotpath="",
        slscolonpath="",
        sls_path="",
        slspath="",
    )


def test_generate_sls_context__one_level_init_implicit():
    """generate_sls_context - Basic one level with implicit init.sls"""
    _test_generated_sls_context(
        "/tmp/foo/init.sls",
        "foo",
        tplfile="foo/init.sls",
        tpldir="foo",
        tpldot="foo",
        slsdotpath="foo",
        slscolonpath="foo",
        sls_path="foo",
        slspath="foo",
    )


def test_generate_sls_context__one_level_init_explicit():
    """generate_sls_context - Basic one level with explicit init.sls"""
    _test_generated_sls_context(
        "/tmp/foo/init.sls",
        "foo.init",
        tplfile="foo/init.sls",
        tpldir="foo",
        tpldot="foo",
        slsdotpath="foo",
        slscolonpath="foo",
        sls_path="foo",
        slspath="foo",
    )


def test_generate_sls_context__one_level():
    """generate_sls_context - Basic one level with name"""
    _test_generated_sls_context(
        "/tmp/foo/boo.sls",
        "foo.boo",
        tplfile="foo/boo.sls",
        tpldir="foo",
        tpldot="foo",
        slsdotpath="foo",
        slscolonpath="foo",
        sls_path="foo",
        slspath="foo",
    )


def test_generate_sls_context__one_level_repeating():
    """generate_sls_context - Basic one level with name same as dir

    (Issue #56410)
    """
    _test_generated_sls_context(
        "/tmp/foo/foo.sls",
        "foo.foo",
        tplfile="foo/foo.sls",
        tpldir="foo",
        tpldot="foo",
        slsdotpath="foo",
        slscolonpath="foo",
        sls_path="foo",
        slspath="foo",
    )


def test_generate_sls_context__two_level_init_implicit():
    """generate_sls_context - Basic two level with implicit init.sls"""
    _test_generated_sls_context(
        "/tmp/foo/bar/init.sls",
        "foo.bar",
        tplfile="foo/bar/init.sls",
        tpldir="foo/bar",
        tpldot="foo.bar",
        slsdotpath="foo.bar",
        slscolonpath="foo:bar",
        sls_path="foo_bar",
        slspath="foo/bar",
    )


def test_generate_sls_context__two_level_init_explicit():
    """generate_sls_context - Basic two level with explicit init.sls"""
    _test_generated_sls_context(
        "/tmp/foo/bar/init.sls",
        "foo.bar.init",
        tplfile="foo/bar/init.sls",
        tpldir="foo/bar",
        tpldot="foo.bar",
        slsdotpath="foo.bar",
        slscolonpath="foo:bar",
        sls_path="foo_bar",
        slspath="foo/bar",
    )


def test_generate_sls_context__two_level():
    """generate_sls_context - Basic two level with name"""
    _test_generated_sls_context(
        "/tmp/foo/bar/boo.sls",
        "foo.bar.boo",
        tplfile="foo/bar/boo.sls",
        tpldir="foo/bar",
        tpldot="foo.bar",
        slsdotpath="foo.bar",
        slscolonpath="foo:bar",
        sls_path="foo_bar",
        slspath="foo/bar",
    )


def test_generate_sls_context__two_level_repeating():
    """generate_sls_context - Basic two level with name same as dir

    (Issue #56410)
    """
    _test_generated_sls_context(
        "/tmp/foo/foo/foo.sls",
        "foo.foo.foo",
        tplfile="foo/foo/foo.sls",
        tpldir="foo/foo",
        tpldot="foo.foo",
        slsdotpath="foo.foo",
        slscolonpath="foo:foo",
        sls_path="foo_foo",
        slspath="foo/foo",
    )


@pytest.mark.skip_on_windows
def test_generate_sls_context__backslash_in_path():
    """generate_sls_context - Handle backslash in path on non-windows"""
    _test_generated_sls_context(
        "/tmp/foo/foo\\foo.sls",
        "foo.foo\\foo",
        tplfile="foo/foo\\foo.sls",
        tpldir="foo",
        tpldot="foo",
        slsdotpath="foo",
        slscolonpath="foo",
        sls_path="foo",
        slspath="foo",
    )
